# pylint: skip-file

import json


TEXT = '''WEAT_DATA 1: We use the flower and insect target words along with pleasant and unpleasant attributes
found in (5).
• Flowers: aster, clover, hyacinth, marigold, poppy, azalea, crocus, iris, orchid, rose, bluebell,
daffodil, lilac, pansy, tulip, buttercup, daisy, lily, peony, violet, carnation, gladiola,
magnolia, petunia, zinnia.
• Insects: ant, caterpillar, flea, locust, spider, bedbug, centipede, fly, maggot, tarantula,
bee, cockroach, gnat, mosquito, termite, beetle, cricket, hornet, moth, wasp, blackfly,
dragonfly, horsefly, roach, weevil.
• Pleasant: caress, freedom, health, love, peace, cheer, friend, heaven, loyal, pleasure, diamond,
gentle, honest, lucky, rainbow, diploma, gift, honor, miracle, sunrise, family,
happy, laughter, paradise, vacation.
• Unpleasant: abuse, crash, filth, murder, sickness, accident, death, grief, poison, stink,
assault, disaster, hatred, pollute, tragedy, divorce, jail, poverty, ugly, cancer, kill, rotten,
vomit, agony, prison.
WEAT_DATA 2: We use the musical instruments and weapons target words along with pleasant and
unpleasant attributes found in (5).
• Instruments: bagpipe, cello, guitar, lute, trombone, banjo, clarinet, harmonica, mandolin,
trumpet, bassoon, drum, harp, oboe, tuba, bell, fiddle, harpsichord, piano, viola, bongo,
flute, horn, saxophone, violin.
• Weapons: arrow, club, gun, missile, spear, axe, dagger, harpoon, pistol, sword, blade,
dynamite, hatchet, rifle, tank, bomb, firearm, knife, shotgun, teargas, cannon, grenade,
mace, slingshot, whip.
• Pleasant: As per previous experiment with insects and flowers.
• Unpleasant: As per previous experiment with insects and flowers.
WEAT_DATA 3: We use the European American and African American names along with pleasant
and unpleasant attributes found in (5). Names that are marked with italics are excluded from
our replication. In the case of African American names this was due to being to infrequent to
occur in GloVe’s Common Crawl corpus; in the case of European American names an equal
number were deleted, chosen at random.
• European American names: Adam, Chip, Harry, Josh, Roger, Alan, Frank, Ian, Justin,
Ryan, Andrew, Fred, Jack, Matthew, Stephen, Brad, Greg, Jed, Paul, Todd, Brandon,
Hank, Jonathan, Peter, Wilbur, Amanda, Courtney, Heather, Melanie, Sara, Amber, Crystal,
Katie, Meredith, Shannon, Betsy, Donna, Kristin, Nancy, Stephanie, Bobbie-Sue,
Ellen, Lauren, Peggy, Sue-Ellen, Colleen, Emily, Megan, Rachel, Wendy.
• African American names: Alonzo, Jamel, Lerone, Percell, Theo, Alphonse, Jerome,
Leroy, Rasaan, Torrance, Darnell, Lamar, Lionel, Rashaun, Tyree, Deion, Lamont, Malik,
Terrence, Tyrone, Everol, Lavon, Marcellus, Terryl, Wardell, Aiesha, Lashelle, Nichelle,
Shereen, Temeka, Ebony, Latisha, Shaniqua, Tameisha, Teretha, Jasmine, Latonya, Shanise,
Tanisha, Tia, Lakisha, Latoya, Sharise, Tashika, Yolanda, Lashandra, Malika, Shavonn,
Tawanda, Yvette.
• Pleasant: caress, freedom, health, love, peace, cheer, friend, heaven, loyal, pleasure, diamond,
gentle, honest, lucky, rainbow, diploma, gift, honor, miracle, sunrise, family,
happy, laughter, paradise, vacation.
• Unpleasant: abuse, crash, filth, murder, sickness, accident, death, grief, poison, stink,
assault, disaster, hatred, pollute, tragedy, bomb, divorce, jail, poverty, ugly, cancer, evil,
kill, rotten, vomit.
WEAT_DATA 4: We use the European American and African American names from (7), along with
pleasant and unpleasant attributes found in (5).
• European American names: Brad, Brendan, Geoffrey, Greg, Brett, Jay, Matthew, Neil,
Todd, Allison, Anne, Carrie, Emily, Jill, Laurie, Kristen, Meredith, Sarah.
• African American names: Darnell, Hakim, Jermaine, Kareem, Jamal, Leroy, Rasheed,
Tremayne, Tyrone, Aisha, Ebony, Keisha, Kenya, Latonya, Lakisha, Latoya, Tamika,
Tanisha.
• Pleasant: caress, freedom, health, love, peace, cheer, friend, heaven, loyal, pleasure, diamond,
gentle, honest, lucky, rainbow, diploma, gift, honor, miracle, sunrise, family,
happy, laughter, paradise, vacation.
• Unpleasant: abuse, crash, filth, murder, sickness, accident, death, grief, poison, stink,
assault, disaster, hatred, pollute, tragedy, bomb, divorce, jail, poverty, ugly, cancer, evil,
kill, rotten, vomit.
WEAT_DATA 5: We use the European American and African American names from (7), along with
pleasant and unpleasant attributes found in (9).
• European American names: Brad, Brendan, Geoffrey, Greg, Brett, Jay, Matthew, Neil,
Todd, Allison, Anne, Carrie, Emily, Jill, Laurie, Kristen, Meredith, Sarah.
• African American names: Darnell, Hakim, Jermaine, Kareem, Jamal, Leroy, Rasheed,
Tremayne, Tyrone, Aisha, Ebony, Keisha, Kenya, Latonya, Lakisha, Latoya, Tamika,
Tanisha.
• Pleasant: joy, love, peace, wonderful, pleasure, friend, laughter, happy.
• Unpleasant: agony, terrible, horrible, nasty, evil, war, awful, failure.
WEAT_DATA 6: We use the male and female names along with career and family attributes found
in (9).
• Male names: John, Paul, Mike, Kevin, Steve, Greg, Jeff, Bill.
• Female names: Amy, Joan, Lisa, Sarah, Diana, Kate, Ann, Donna.
• Career: executive, management, professional, corporation, salary, office, business, career.
• Family: home, parents, children, family, cousins, marriage, wedding, relatives.
WEAT_DATA 7: We use the math and arts target words along with male and female attributes found
in (9).
• Math: math, algebra, geometry, calculus, equations, computation, numbers, addition.
• Arts: poetry, art, dance, literature, novel, symphony, drama, sculpture.
• Male terms: male, man, boy, brother, he, him, his, son.
• Female terms: female, woman, girl, sister, she, her, hers, daughter.
WEAT_DATA 8: We use the science and arts target words along with male and female attributes
found in (10).
• Science: science, technology, physics, chemistry, Einstein, NASA, experiment, astronomy.
• Arts: poetry, art, Shakespeare, dance, literature, novel, symphony, drama.
• Male terms: brother, father, uncle, grandfather, son, he, his, him.
• Female terms: sister, mother, aunt, grandmother, daughter, she, hers, her.
WEAT_DATA 9: We use the mental and physical disease target words along with uncontrollability
and controllability attributes found in (23).
• Mental disease: sad, hopeless, gloomy, tearful, miserable, depressed.
• Physical disease: sick, illness, influenza, disease, virus, cancer.
• Temporary: impermanent, unstable, variable, fleeting, short-term, brief, occasional.
• Permanent: stable, always, constant, persistent, chronic, prolonged, forever.
WEAT_DATA 10: We use young and old people’s names as target words along with pleasant and
unpleasant attributes found in (9).
• Young people’s names: Tiffany, Michelle, Cindy, Kristy, Brad, Eric, Joey, Billy.
• Old people’s names: Ethel, Bernice, Gertrude, Agnes, Cecil, Wilbert, Mortimer, Edgar.
• Pleasant: joy, love, peace, wonderful, pleasure, friend, laughter, happy.
• Unpleasant: agony, terrible, horrible, nasty, evil, war, awful, failure.'''


def parse_line(line):
    name, words_str = line[1:].split(': ')
    return {'name': name, 'words': words_str[:-2].split(', ')}


def parse_case(case):
    groups_str = case.replace('\n', ' ').split('•')[1:]
    return {'first_target': parse_line(groups_str[0]),
            'second_target': parse_line(groups_str[1]),
            'first_attribute': parse_line(groups_str[2]),
            'second_attribute': parse_line(groups_str[3])}


cases = TEXT.split('WEAT_DATA')[1:]

WEAT_DATA = [parse_case(case) for case in cases]

WEAT_DATA[1]['first_attribute']['words'] = WEAT_DATA[0]['first_attribute']['words']
WEAT_DATA[1]['second_attribute']['words'] = WEAT_DATA[0]['second_attribute']['words']

WEAT_DATA[2]['first_target']['remove'] = ['Chip', 'Ian', 'Fred', 'Jed', 'Todd', 'Brandon', 'Hank', 'Wilbur', 'Sara', 'Amber', 'Crystal', 'Meredith', 'Shannon', 'Donna', 'Bobbie-Sue', 'Peggy', 'Sue-Ellen', 'Wendy']
WEAT_DATA[2]['second_target']['remove'] = ['Lerone', 'Percell', 'Rasaan', 'Rashaun', 'Everol', 'Terryl', 'Aiesha', 'Lashelle', 'Temeka', 'Tameisha', 'Teretha', 'Latonya', 'Shanise', 'Sharise', 'Tashika', 'Lashandra', 'Shavonn', 'Tawanda']

assert len(WEAT_DATA[2]['first_target']['remove']) == len(WEAT_DATA[2]['second_target']['remove'])
assert set(WEAT_DATA[2]['first_target']['remove']).issubset(WEAT_DATA[2]['first_target']['words'])

WEAT_DATA[3]['first_target']['remove'] = ['Jay', 'Kristen']
WEAT_DATA[3]['second_target']['remove'] = ['Tremayne', 'Latonya']

assert len(WEAT_DATA[3]['first_target']['remove']) == len(WEAT_DATA[3]['second_target']['remove'])
assert set(WEAT_DATA[3]['first_target']['remove']).issubset(WEAT_DATA[3]['first_target']['words'])

WEAT_DATA[4]['first_target']['remove'] = ['Jay', 'Kristen']
WEAT_DATA[4]['second_target']['remove'] = ['Tremayne', 'Latonya']

assert len(WEAT_DATA[4]['first_target']['remove']) == len(WEAT_DATA[4]['second_target']['remove'])
assert set(WEAT_DATA[4]['first_target']['remove']).issubset(WEAT_DATA[4]['first_target']['words'])


WEAT_DATA[0]['original_finding'] = {'Ref': 'A. G. Greenwald, D. E. McGhee, J. L. Schwartz, Measuring individual differences in im- plicit cognition: the implicit association test., Journal of personality and social psychology 74, 1464 (1998).',
                                    'N': '32',
                                    'd': '1.35',
                                    'p': '1e-8'}

WEAT_DATA[1]['original_finding'] = {'Ref': 'A. G. Greenwald, D. E. McGhee, J. L. Schwartz, Measuring individual differences in im- plicit cognition: the implicit association test., Journal of personality and social psychology 74, 1464 (1998).',
                                    'N': '32',
                                    'd': '1.66',
                                    'p': '1e-10'}

WEAT_DATA[2]['original_finding'] = {'Ref': 'A. G. Greenwald, D. E. McGhee, J. L. Schwartz, Measuring individual differences in im- plicit cognition: the implicit association test., Journal of personality and social psychology 74, 1464 (1998).',
                                    'N': '26',
                                    'd': '1.17',
                                    'p': '1e-5'}

WEAT_DATA[3]['original_finding'] = {'Ref': 'M. Bertrand, S. Mullainathan, Are Emily and Greg more employable than Lakisha and Jamal? a field experiment on labor market discrimination, The American Economic Review 94, 991 (2004).',
                                    'N': '',
                                    'd': '',
                                    'p': ''}

WEAT_DATA[4]['original_finding'] = {'Ref': 'M. Bertrand, S. Mullainathan, Are Emily and Greg more employable than Lakisha and Jamal? a field experiment on labor market discrimination, The American Economic Review 94, 991 (2004).',
                                    'N': '',
                                    'd': '',
                                    'p': ''}

WEAT_DATA[5]['original_finding'] = {'Ref': 'B. A. Nosek, M. Banaji, A. G. Greenwald, Harvesting implicit group attitudes and beliefs from a demonstration web site., Group Dynamics: Theory, Research, and Practice 6, 101 (2002).',
                                    'N': '39k',
                                    'd': '0.72',
                                    'p': '< 1e-2'}

WEAT_DATA[6]['original_finding'] = {'Ref': 'B. A. Nosek, M. Banaji, A. G. Greenwald, Harvesting implicit group attitudes and beliefs from a demonstration web site., Group Dynamics: Theory, Research, and Practice 6, 101 (2002).',
                                    'N': '28k',
                                    'd': '0.82',
                                    'p': '< 1e-2'}

WEAT_DATA[7]['original_finding'] = {'Ref': 'B. A. Nosek, M. R. Banaji, A. G. Greenwald, Math=male, me=female, therefore math̸=me., Journal of Personality and Social Psychology 83, 44 (2002).',
                                    'N': '91',
                                    'd': '1.47',
                                    'p': '1e-24'}

WEAT_DATA[8]['original_finding'] = {'Ref': 'P. D. Turney, P. Pantel, From frequency to meaning: Vector space models of semantics, Journal of Artificial Intelligence Research 37, 141 (2010).',
                                    'N': '135',
                                    'd': '1.01',
                                    'p': '1e-3'}


WEAT_DATA[9]['original_finding'] = {'Ref': 'B. A. Nosek, M. Banaji, A. G. Greenwald, Harvesting implicit group attitudes and beliefs from a demonstration web site., Group Dynamics: Theory, Research, and Practice 6, 101 (2002).',
                                    'N': '43k',
                                    'd': '1.42',
                                    'p': '< 1e-2'}

json.dump(WEAT_DATA, open('weat.json', 'w'), indent=True)



"""
Algorithmic intervensions for satisfying fairness criteria.

There are three families of techniques:

1. **Pre-processing** - Adjust features in the dataset.
2. **In-processing** - Adjust the learning algorithm.
3. **Post-processing** - Adjust the learned classifier.

"""



from collections import Counter
from functools import partial

import numpy as np
import pandas as pd
from sklearn.metrics import roc_auc_score, roc_curve
from sklearn.utils.multiclass import unique_labels

from responsibly.fairness.metrics.utils import (
    _assert_binary, _groupby_y_x_sens,
)


def _proportion(data, labels):
    counts = Counter(data)
    assert set(counts.keys()).issubset(labels)
    return (counts[labels[1]]
            / (counts[labels[0]] + counts[labels[1]]))


def _get_labels(ys, labels):

    if labels is None:
        labels = unique_labels(ys)
    else:
        labels = np.asarray(labels)
        if np.all([label not in ys for label in labels]):
            raise ValueError('At least one label specified must be in y.')

    return labels


def _normalize_by_attr(y_score, x_sens, ndigits=1):
    y_score_within = y_score[:]

    for indices in x_sens.groupby(x_sens).groups.values():
        y_score_within[indices] = (y_score_within[indices]
                                   .rank(pct=True))

    y_score_within = (np.floor(y_score_within * (10**ndigits))
                      / (10**ndigits))

    return y_score_within


def independence_score(y_score, x_sens,
                       as_df=False):
    """Compute the independence criteria for score prediction.

    In classification terminology, it is the **acceptance rate**
    grouped by the score and the sensitive attribute.

    :param y_score: Estimated target score as returned by a classifier.
    :param x_sens: Sensitive attribute values corresponded to each
                   estimated target.
    :param as_df: Whether to return the results as ``dict`` (if ``False``)
                  or as :class:`pandas.DataFrame`(if ``True``).
    :return: Independence criteria.
    :rtype: dict or :class:`pandas.DataFrame`
    """
    criterion = pd.crosstab(index=y_score,
                            columns=x_sens,
                            normalize='columns')

    if not as_df:
        criterion = criterion.to_dict()

    return criterion


def separation_score(y_true, y_score, x_sens,
                     labels=None,
                     as_df=False):
    """Compute the separation criteria for score prediction.

    In classification terminology, it is the **FPR** and **TPR**
    grouped by the score and the sensitive attribute.

    :param y_true: Binary ground truth (correct) target values.
    :param y_score: Estimated target score as returned by a classifier.
    :param x_sens: Sensitive attribute values corresponded to each
                   estimated target.
    :param as_df: Whether to return the results as ``dict`` (if ``False``)
                  or as :class:`pandas.DataFrame` (if ``True``).
    :return: Separation criteria.
    :rtype: dict or :class:`pandas.DataFrame`
    """

    _assert_binary(y_true)

    labels = _get_labels(y_score, labels)

    criterion = pd.crosstab(index=y_score,
                            columns=[y_true, x_sens],
                            normalize=True)

    if not as_df:
        criterion = criterion.to_dict()

    return criterion


def sufficiency_score(y_true, y_score, x_sens,
                      labels=None,
                      within_score_percentile=False,
                      as_df=False):
    """Compute the sufficiency criteria for score prediction.

    In classification terminology, it is the **PPV** and the **NPV**
    grouped by the score and the sensitive attribute.

    :param y_true: Binary ground truth (correct) target values.
    :param y_score: Estimated target score as returned by a classifier.
    :param x_sens: Sensitive attribute values corresponded to each
                   target.
    :param as_df: Whether to return the results as ``dict`` (if ``False``)
                  or as :class:`pandas.DataFrame` (if ``True``).
    :return: Sufficiency criteria.
    :rtype: dict or :class:`pandas.DataFrame`
    """

    _assert_binary(y_true)

    labels = _get_labels(y_true, labels)

    if within_score_percentile:
        y_score = _normalize_by_attr(y_score, x_sens,
                                     within_score_percentile)

    criterion = pd.crosstab(index=y_score,
                            columns=x_sens,
                            values=y_true,
                            aggfunc=partial(_proportion,
                                            labels=labels))

    if not as_df:
        criterion = criterion.to_dict()

    return criterion


def _all_equal(iterator):
    iterator = iter(iterator)

    try:
        first = next(iterator)
    except StopIteration:
        return True

    try:
        return all(np.allclose(first, rest) for rest in iterator)
    except ValueError:
        return False


def roc_curve_by_attr(y_true, y_score, x_sens,
                      pos_label=None, sample_weight=None,
                      drop_intermediate=False):
    """Compute Receiver operating characteristic (ROC) by attribute.

    Based on :func:`sklearn.metrics.roc_curve`

    :param y_true: Binary ground truth (correct) target values.
    :param y_score: Estimated target score as returned by a classifier.
    :param x_sens: Sensitive attribute values corresponded to each
                   estimated target.
    :param pos_label: Label considered as positive and others
                      are considered negative.
    :param sample_weight: Sample weights.
    :param drop_intermediate: Whether to drop some suboptimal
                              thresholds which would not appear on
                              a plotted ROC curve.
                              This is useful in order to create
                              lighter ROC curves.
    :return: For each value of sensitive attribute:
             - fpr - Increasing false positive rates such
               that element i is the false positive rate
               of predictions with score >= thresholds[i].
             - fpr - Increasing true positive rates such
               that element i is the true positive rate
               of predictions with score >= thresholds[i].
             - thresholds -
               Decreasing thresholds on the decision function
               used to compute fpr and tpr. thresholds[0] represents
               no instances being predicted and is arbitrarily set
               to max(y_score) + 1.
    :rtype: dict

    """

    grouped = _groupby_y_x_sens(y_true, y_score, x_sens)

    # pylint: disable=too-many-function-args
    roc_curves = {x_sens_value: roc_curve(group['y_true'],
                                          group['y_score'],
                                          pos_label, sample_weight,
                                          drop_intermediate)
                  for x_sens_value, group in grouped}

    if not _all_equal(thresholds
                      for _, _, thresholds in roc_curves.values()):
        raise NotImplementedError('All the scores values should'
                                  ' appear for each sensitive'
                                  ' attribute value.'
                                  ' It will be implemented'
                                  ' in the future.'
                                  ' Please post your use-case in'
                                  ' https://github.com/ResponsiblyAI/responsibly/issues/15')  # pylint: disable=line-too-long

    return roc_curves


def roc_auc_score_by_attr(y_true, y_score, x_sens,
                          sample_weight=None):
    """Compute Area Under the ROC (AUC) by attribute.

    Based on function:`sklearn.metrics.roc_auc_score`

    :param y_true: Binary ground truth (correct) target values.
    :param y_score: Estimated target score as returned by a classifier.
    :param x_sens: Sensitive attribute values corresponded to each
                   estimated target.
    :param sample_weight: Sample weights.
    :return: ROC AUC grouped by the sensitive attribute.
    :rtype: dict
    """

    grouped = _groupby_y_x_sens(y_true, y_score, x_sens)

    return {x_sens_value: roc_auc_score(group['y_true'],
                                        group['y_score'],
                                        sample_weight=sample_weight)
            for x_sens_value, group in grouped}



import pandas as pd
from pandas.core.algorithms import unique as _unique
from sklearn.metrics import confusion_matrix

from responsibly.fairness.metrics.utils import _assert_binary


def _select_dict(d, keys):
    return {k: d[k] for k in keys}


def _nested_select_dict(d, nested_keys):
    return {k:
            _select_dict(v, nested_keys)
            for k, v in d.items()}


def _choose_other(item, iterable):
    return next(other for other in iterable
                if other != item)


def _nested_diff_and_ratio(d, nested_key, first, second):

    assert d.keys() == {first, second}

    return {'diff': d[first][nested_key] - d[second][nested_key],
            'ratio': d[first][nested_key] / d[second][nested_key]}


def binary_stats_by_attr(y_true, y_pred, x_attr,
                         labels=None):
    # pylint: disable=too-many-locals

    _assert_binary(y_true, y_pred)

    stats = {}

    for x_att_val in _unique(x_attr):
        mask = (x_attr == x_att_val)

        tn, fp, fn, tp = confusion_matrix(y_true[mask],
                                          y_pred[mask],
                                          labels=labels).ravel()

        pos = tp + fn
        neg = tn + fp

        acceptance = tp + fp
        rejection = tn + fn

        correct = tp + tn

        total = pos + neg

        stats[x_att_val] = {
            'total': int(total),
            'proportion': total / len(x_attr),
            'pos': int(pos),
            'neg': int(neg),
            'base_rate': pos / total,
            'acceptance_rate': acceptance / total,
            'tn': int(tn),
            'fp': int(fp),
            'fn': int(fn),
            'tp': int(tp),
            'accuracy': correct / total,
            'balanced_accuracy': (tp / pos + tn / neg) / 2,
            'tpr': tp / pos,
            'tnr': tn / neg,
            'fnr': fn / pos,
            'fpr': fp / neg,
            'ppv': tp / acceptance,
            'npv': tn / rejection
        }

    return stats


def compare_privileged(stats,
                       x_sens_privileged=None):
    # pylint: disable=line-too-long

    if len(stats) != 2:
        if x_sens_privileged is not None:
            raise ValueError('x_sens_privileged should have'
                             'only two values for comparision'
                             '(difference and ratio).')

        return None

    comparison = {}

    if x_sens_privileged is None:
        x_sens_privileged = next(iter(stats))

    x_sens_unprivileged = _choose_other(x_sens_privileged,
                                        stats)

    comparison['x_sens_privileged'] = x_sens_privileged
    comparison['x_sens_unprivileged'] = x_sens_unprivileged

    comparison['metrics'] = {}

    metrics = next(iter(stats.values())).keys()

    for metric in metrics:
        comparison['metrics'][metric] = _nested_diff_and_ratio(stats,
                                                               metric,
                                                               x_sens_unprivileged,
                                                               x_sens_privileged)

    return comparison


def group_fairness_criterion_binary(y_true, y_pred, x_sens,
                                    metrics,
                                    x_sens_privileged=None,
                                    labels=None,
                                    as_df=False):

    stats = binary_stats_by_attr(y_true, y_pred, x_sens,
                                 labels=labels)

    criterion = _nested_select_dict(stats,
                                    metrics)

    comparison = compare_privileged(criterion,
                                    x_sens_privileged)

    if as_df:
        criterion = pd.DataFrame(criterion)

        if comparison is not None:
            vs_name = ('{x_sens_unprivileged} vs. {x_sens_privileged}'
                       .format(**comparison))

            comparison = pd.DataFrame(comparison['metrics'])
            comparison.index.name = vs_name

    return criterion, comparison


def independence_binary(y_pred, x_sens,
                        x_sens_privileged=None,
                        labels=None,
                        as_df=False):
    """Compute the independence criteria for binary prediction.

    In classification terminology, it is the **acceptance rate**
    grouped by the sensitive attribute.

    :param y_pred: Estimated targets as returned by a classifier.
    :param x_sens: Sensitive attribute values corresponded to each
                   target.
    :param x_sens_privileged: The privileged value in the
                              sensitive attribute. Relevent only
                              if there are only two values for
                              the sensitive attribute.
    :param labels: List of labels to choose the negative and positive target.
                   This may be used to reorder or select a subset of labels.
                   If none is given, those that appear at least once in
                   y_pred are used in sorted order; first is negative
                   and the second is positive.
    :param as_df: Whether to return the results as `dict` (if `False`)
                  or as :class:`pandas.DataFrame` (if `True`).
    :return: Independence criteria and comparision if there are
             only two values for the sensitive attribute.
    :rtype: tuple
    """

    # hack to keep the same strutcure of code
    # for independence as seperation and sufficiency
    # we take only acceptance_rate
    return group_fairness_criterion_binary(y_pred, y_pred, x_sens,
                                           ('acceptance_rate',),
                                           x_sens_privileged,
                                           labels,
                                           as_df)


def separation_binary(y_true, y_pred, x_sens,
                      x_sens_privileged=None,
                      labels=None,
                      as_df=False):
    """Compute the separation criteria for binary prediction.

    In classification terminology, it is the **TPR**, **FPR**,
    **TNR** and **FNR** grouped by the sensitive attribute.

    :param y_true: Binary ground truth (correct) target values.
    :param y_pred: Estimated binary targets as returned
                   by a classifier.
    :param x_sens: Sensitive attribute values corresponded to each
                   target.
    :param x_sens_privileged: The privileged value in the
                              sensitive attribute. Relevent only
                              if there are only two values for
                              the sensitive attribute.
    :param labels: List of labels to choose the negative and positive target.
                   This may be used to reorder or select a subset of labels.
                   If none is given, those that appear at least once in
                   y_pred are used in sorted order; first is negative
                   and the second is positive.
    :param as_df: Whether to return the results as `dict` (if `False`)
                  or as :class:`pandas.DataFrame` (if `True`).
    :return: Separation criteria and comparision if there are
             only two values for the sensitive attribute.
    :rtype: tuple
    """

    return group_fairness_criterion_binary(y_true, y_pred, x_sens,
                                           ('tpr', 'fpr', 'tnr', 'fnr'),
                                           x_sens_privileged,
                                           labels,
                                           as_df)


def sufficiency_binary(y_true, y_pred, x_sens,
                       x_sens_privileged=None,
                       labels=None,
                       as_df=False):
    """Compute the sufficiency criteria for binary prediction.

    In classification terminology, it is the **PPV** and **NPV**
    grouped by the sensitive attribute.

    :param y_true: Binary ground truth (correct) target values.
    :param y_pred: Binary estimated targets as returned by
                   a classifier.
    :param x_sens: Sensitive attribute values corresponded to each
                   target.
    :param x_sens_privileged: The privileged value in the
                              sensitive attribute. Relevent only
                              if there are only two values for
                              the sensitive attribute.
    :param labels: List of labels to choose the negative and positive target.
                   This may be used to reorder or select a subset of labels.
                   If none is given, those that appear at least once in
                   y_pred are used in sorted order; first is negative
                   and the second is positive.
    :param as_df: Whether to return the results as `dict` (if `False`)
                  or as :class:`pandas.DataFrame` (if `True`).
    :return: Sufficiency criteria and comparision if there are
             only two values for the sensitive attribute.
    :rtype: tuple
    """

    return group_fairness_criterion_binary(y_true, y_pred, x_sens,
                                           ('ppv', 'npv'),
                                           x_sens_privileged,
                                           labels,
                                           as_df)


def report_binary(y_true, y_pred, x_sens,
                  labels=None):
    """Generate a report of criteria for binary prediction.

    In classification terminology, the statistics are
    grouped by the sensitive attribute:
    - Number of observations per group
    - Proportion of of observations per group
    - Base rate
    - Acceptance rate
    - FNR
    - TPR
    - PPV
    - NPV

    :param y_true: Binary ground truth (correct) target values.
    :param y_pred: Binary estimated targets as returned by
                   a classifier.
    :param x_sens: Sensitive attribute values corresponded to each
                   target.
    :param labels: List of labels to choose the negative and positive target.
                   This may be used to reorder or select a subset of labels.
                   If none is given, those that appear at least once in
                   y_pred are used in sorted order; first is negative
                   and the second is positive.
    :return: Classification statistics grouped by the
             sensitive attribute.
    :rtype: :class:`pandas.DataFrame`
    """

    stats = binary_stats_by_attr(y_true, y_pred, x_sens, labels)
    stats_df = pd.DataFrame(stats)

    return stats_df.loc[['total', 'proportion', 'base_rate',
                         'acceptance_rate', 'accuracy',
                         'fnr', 'fpr', 'ppv', 'npv']]



"""
Collection of common benchmark datasets from fairness research.

Each dataset object contains a :class:`pandas.DataFrame` as `df` attribute
that holds the actual data.
The dataset object will take care of loading, preprocessing
and validating the data.
The preprocessing is done by standard practices that are associated with
this data set: from its manual (e.g., README)
or as other did in the literature.

See :class:`responsibly.dataset.Dataset`
for additional attribute and complete documentation.

Currently these are the available datasets:

    - ProPublica recidivism/COMPAS dataset,
      see: :class:`~responsibly.dataset.COMPASDataset`

    - Adult dataset,
      see: :class:`~responsibly.dataset.AdultDataset`

    - German credit dataset,
      see: :class:`~responsibly.dataset.GermanDataset`

    - FICO credit score dataset,
      see :func:`~responsibly.dataset.build_FICO_dataset`

Usage
-----
.. code:: python

    >>> from responsibly.dataset import COMPASDataset
    >>> compas_ds = COMPASDataset()
    >>> print(compas_ds)
    <ProPublica Recidivism/COMPAS Dataset. 6172 rows, 56 columns in
    which {race, sex} are sensitive attributes>
    >>> type(compas_ds.df)
    <class 'pandas.core.frame.DataFrame'>
    >>> compas_ds.df['race'].value_counts()
    African-American    3175
    Caucasian           2103
    Hispanic             509
    Other                343
    Asian                 31
    Native American       11
    Name: race, dtype: int64
"""

from responsibly.dataset.adult import AdultDataset
from responsibly.dataset.compas import COMPASDataset
from responsibly.dataset.core import Dataset
from responsibly.dataset.fico import build_FICO_dataset
from responsibly.dataset.german import GermanDataset



"""Unit tests configuration file."""


def pytest_configure(config):
    """Disable verbose output when running tests."""

    terminal = config.pluginmanager.getplugin('terminal')

    class QuietReporter(terminal.TerminalReporter):
        @property
        def verbosity(self):
            return 0

        @property
        def showlongtestinfo(self):
            return False

        @property
        def showfspath(self):
            return False

    terminal.TerminalReporter = QuietReporter



# TODO how import files from a package
import json
import warnings

from gensim.models.keyedvectors import KeyedVectors
from pkg_resources import resource_filename, resource_string


def load_w2v_small():
    """Load reduced Word2Vec model as `KeyedVectors` object.

    Based on the pre-trained embedding on the Google News corpus:
    https://code.google.com/archive/p/word2vec/
    """
    # pylint: disable=C0301

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', DeprecationWarning)
        model = KeyedVectors.load_word2vec_format(
            resource_filename(__name__, 'GoogleNews-vectors-negative300-bolukbasi.bin'),
            binary=True)

    return model


def load_json_resource(resource_name):
    return json.loads(
        resource_string(__name__, resource_name + '.json').decode('utf-8')

    )


BOLUKBASI_DATA = load_json_resource('bolukbasi')

BOLUKBASI_DATA['gender']['profession_names'] = list(
    zip(*BOLUKBASI_DATA['gender']['professions']))[0]


BOLUKBASI_DATA['gender']['specific_full'].sort()

# TODO: in the code of the article, the last definitional pair
# is not in the specific full
BOLUKBASI_DATA['gender']['specific_full_with_definitional_equalize'] = list(
    (set.union(
        *map(set, BOLUKBASI_DATA['gender']['definitional_pairs']))
     | set.union(
         *map(set, BOLUKBASI_DATA['gender']['equalize_pairs']))
     | set(BOLUKBASI_DATA['gender']['specific_full']))
)
BOLUKBASI_DATA['gender']['specific_full_with_definitional_equalize'].sort()

BOLUKBASI_DATA['gender']['neutral_profession_names'] = list(
    set(BOLUKBASI_DATA['gender']['profession_names'])
    - set(BOLUKBASI_DATA['gender']['specific_full_with_definitional_equalize'])
)
BOLUKBASI_DATA['gender']['neutral_profession_names'].sort()

BOLUKBASI_DATA['gender']['word_group_keys'] = ['profession_names',
                                               'neutral_profession_names',
                                               'specific_seed',
                                               'specific_full',
                                               'specific_full_with_definitional_equalize']  # pylint: disable=C0301


WEAT_DATA = load_json_resource('weat')

# Zhao, J., Wang, T., Yatskar, M., Ordonez, V., & Chang, K. W. (2018).
# Gender bias in coreference resolution: Evaluation and debiasing methods.
# arXiv preprint arXiv:1804.06876.
# https://arxiv.org/abs/1804.06876
OCCUPATION_FEMALE_PRECENTAGE = load_json_resource(
    'occupational_female_precentage')



"""
Demographic fairness in binary classification.

The common and simples setting, but not the only one, of fairness
of a binary classifier is the demographic one.
It is assume that there is one sensitive attribute or more
that represents one or more demographic groups
(e.g., by gender, race or age), for which a classifier
should be fair.

.. important::
    The terminology and functionality is aligned with the
    book `Fairness and Machine Learning
    - Limitations and Opportunities <https://fairmlbook.org>`_
    by Solon Barocas, Moritz Hardt and Arvind Narayanan.
    Therefore, it is advised to get familiar with
    `Chapter 2 <https://fairmlbook.org/demographic.html>`_,
    as it summarized the current core knowledge regarding fairness
    in classification.

Currently, the :mod:`responsibly.fairness` module has two components:

1. **Metrics** (:mod:`responsibly.fairness.metrics`) for measuring unfairness.
2. **Algorithmic interventions** (:mod:`responsibly.fairness.interventions`)
   for satisfying fairness criteria.

"""



__all__ = ['COMPASDataset']

import numpy as np
import pandas as pd
from pkg_resources import resource_filename

from responsibly.dataset.core import Dataset


COMPAS_PATH = resource_filename(__name__,
                                'compas-scores-two-years.csv')


class COMPASDataset(Dataset):
    """ProPublica Recidivism/COMPAS Dataset.

    See :class:`~responsibly.dataset.Dataset` for a description of
    the arguments and attributes.

    References:
        https://github.com/propublica/compas-analysis

    """

    def __init__(self):
        super().__init__(target='is_recid',
                         sensitive_attributes=['race', 'sex'],
                         prediction=['y_pred', 'score_factor',
                                     'score_text'])

    def _load_data(self):
        return pd.read_csv(COMPAS_PATH)

    def _preprocess(self):
        """Perform the same preprocessing as the original analysis.

        https://github.com/propublica/compas-analysis/blob/master/Compas%20Analysis.ipynb
        """

        self.df = self.df[(self.df['days_b_screening_arrest'] <= 30)
                          & (self.df['days_b_screening_arrest'] >= -30)
                          & (self.df['is_recid'] != -1)
                          & (self.df['c_charge_degree'] != 'O')
                          & (self.df['score_text'] != 'N/A')]

        self.df['c_jail_out'] = pd.to_datetime(self.df['c_jail_out'])
        self.df['c_jail_in'] = pd.to_datetime(self.df['c_jail_in'])
        self.df['length_of_stay'] = (self.df['c_jail_out']
                                     - self.df['c_jail_in'])

        self.df['score_factor'] = np.where(self.df['score_text']
                                           != 'Low',
                                           'HighScore', 'LowScore')
        self.df['y_pred'] = (self.df['score_factor'] == 'HighScore')

    def _validate(self):
        # pylint: disable=line-too-long
        super()._validate()

        assert len(self.df) == 6172, 'the number of rows should be 6172,'\
                                     ' but it is {}.'.format(len(self.df))
        assert len(self.df.columns) == 56, 'the number of columns should be 56,'\
                                           ' but it is {}.'.format(len(self.df.columns))



__all__ = ['AdultDataset']

import numpy as np
import pandas as pd
from pkg_resources import resource_filename

from responsibly.dataset.core import Dataset


ADULT_TRAIN_PATH = resource_filename(__name__,
                                     'adult.data')
ADULT_TEST_PATH = resource_filename(__name__,
                                    'adult.test')

COLUMN_NAMES = ['age', 'workclass', 'fnlwgt', 'education',
                'education-num', 'marital_status', 'occupation',
                'relationship', 'race', 'sex', 'capital_gain',
                'capital_loss', 'hours_per_week', 'native_country',
                'income_per_year']


class AdultDataset(Dataset):
    """Adult Dataset.

    See :class:`~responsibly.dataset.Dataset` for a description of
    the arguments and attributes.

    References:
        https://archive.ics.uci.edu/ml/datasets/adult

    """

    def __init__(self):
        super().__init__(target='income_per_year',
                         sensitive_attributes=['sex', 'race'])

    def _load_data(self):
        train_df = pd.read_csv(ADULT_TRAIN_PATH, names=COLUMN_NAMES,
                               skipinitialspace=True,
                               header=None, index_col=False,
                               na_values='?')
        test_df = pd.read_csv(ADULT_TEST_PATH, names=COLUMN_NAMES,
                              skipinitialspace=True,
                              header=0, index_col=False,
                              na_values='?')

        train_df['dataset'] = 'train'
        test_df['dataset'] = 'test'

        return pd.concat([train_df, test_df], ignore_index=True)

    def _preprocess(self):
        """Perform the same preprocessing as the dataset doc file."""
        self.df = self.df.dropna()
        self.df = self.df.drop(['fnlwgt'], axis=1)
        self.df['income_per_year'] = (self.df['income_per_year']
                                      .str
                                      .replace('.', ''))

    def _validate(self):
        # pylint: disable=line-too-long
        super()._validate()

        assert len(self.df) == 45222, 'the number of rows should be 45222,'\
                                      ' but it is {}.'.format(len(self.df))
        assert len(self.df.columns) == 15, 'the number of columns should be 15,'\
                                           ' but it is {}.'.format(len(self.df.columns))

        train_df = self.df[self.df['dataset'] == 'train']
        test_df = self.df[self.df['dataset'] == 'test']
        assert len(train_df) == 30162, 'the number of train rows should be 30162,'\
                                       ' but it is {}.'.format(len(train_df))
        assert len(test_df) == 15060, 'the number of train rows should be 15060,'\
                                      ' but it is {}.'.format(len(test_df))



"""Unit test module for responsibly.we"""
# pylint: disable=redefined-outer-name,unused-variable,expression-not-assigned,singleton-comparison,protected-access

import copy
from math import isclose

import numpy as np
import pytest

from responsibly.consts import RANDOM_STATE
from responsibly.tests.data import TOLGA_GENDER_ANALOGIES
from responsibly.tests.utils import assert_deep_almost_equal
from responsibly.we import (
    GenderBiasWE, calc_all_weat, calc_weat_pleasant_unpleasant_attribute,
)
from responsibly.we.data import WEAT_DATA, load_w2v_small
from responsibly.we.utils import (
    most_similar, normalize, project_params, project_reject_vector,
    project_vector,
)


ATOL = 1e-6

N_RANDOM_NEUTRAL_WORDS_DEBIAS_TO_TEST = 1000


@pytest.fixture
def w2v_small():
    return load_w2v_small()


@pytest.fixture
def gender_biased_w2v_small():
    model = load_w2v_small()
    return GenderBiasWE(model, only_lower=True, verbose=True)


def test_identify_direction_single(w2v_small):
    gb = GenderBiasWE(w2v_small, only_lower=True, verbose=True,
                      identify_direction='single')

    direction = normalize(w2v_small['she'] - w2v_small['he'])

    np.testing.assert_allclose(gb.direction, direction)


# TODO: only check that there is no exception,
# should b change to a better test cas
def test_identify_direction_sum(w2v_small):
    gb = GenderBiasWE(w2v_small, only_lower=True, verbose=True,
                      identify_direction='sum')


def test_assert_gensim_keyed_vectors():
    with pytest.raises(TypeError):
        GenderBiasWE(['one', 'two'], only_lower=True, verbose=True)


def test_project_params():
    # pylint: disable=arguments-out-of-order
    v = np.array([1, 2, 3])
    u = np.array([-4, 5, -6])

    (_,
     projected_vector_v1,
     rejected_vector_v1) = project_params(v, u)

    projected_vector_v2, rejected_vector_v2 = project_reject_vector(u, v)

    np.testing.assert_allclose(projected_vector_v1, projected_vector_v2)
    np.testing.assert_allclose(rejected_vector_v1, rejected_vector_v2)


def test_words_embbeding_loading(gender_biased_w2v_small):
    assert len(gender_biased_w2v_small.model.vocab) == 26423


def test_contains(gender_biased_w2v_small):
    assert 'home' in gender_biased_w2v_small
    assert 'HOME' not in gender_biased_w2v_small


def test_data_is_sorted_list(gender_biased_w2v_small):
    # otherwise 'specific_full_with_definitional_equalize' is not sorted
    assert gender_biased_w2v_small.only_lower

    for key in gender_biased_w2v_small._data['word_group_keys']:
        word_list = gender_biased_w2v_small._data[key]
        assert isinstance(word_list, list)
        assert all(word_list[i] <= word_list[i + 1]
                   for i in range(len(word_list) - 1))


def test_calc_direct_bias(gender_biased_w2v_small):
    """
    Test calc_direct_bias method in GenderBiasWE.

    Based on section 5.2
    """

    # TODO: it seemse that in the article it was checked on
    # all the professions names including gender specific ones
    # (e.g. businesswomen)
    assert isclose(gender_biased_w2v_small.calc_direct_bias(),
                   0.07, abs_tol=1e-2)
    assert isclose(gender_biased_w2v_small.calc_direct_bias(gender_biased_w2v_small  # pylint: disable=C0301
                                                            ._data['profession_names']),  # pylint: disable=C0301

                   0.08, abs_tol=1e-2)


# TODO: iterate over a dictionary
def test_calc_indirect_bias(gender_biased_w2v_small, all_zero=False):
    """
    Test calc_direct_bias method in GenderBiasWE.

    Based on figure 3 & section 3.5
    """
    assert isclose(gender_biased_w2v_small.calc_indirect_bias('softball',
                                                              'pitcher'),
                   0 if all_zero else -0.01, abs_tol=1e-2)
    assert isclose(gender_biased_w2v_small.calc_indirect_bias('softball',
                                                              'bookkeeper'),
                   0 if all_zero else 0.20, abs_tol=1e-2)
    assert isclose(gender_biased_w2v_small.calc_indirect_bias('softball',
                                                              'receptionist'),
                   0 if all_zero else 0.67, abs_tol=1e-2)
    # these words have legit gender direction projection
    if not all_zero:
        assert isclose(gender_biased_w2v_small.calc_indirect_bias('softball',
                                                                  'registered_nurse'),  # pylint: disable=C0301
                       0 if all_zero else 0.29, abs_tol=1e-2)
        # TODO: in the article it is 0.35 - why?
        assert isclose(gender_biased_w2v_small.calc_indirect_bias('softball',
                                                                  'waitress'),
                       0 if all_zero else 0.31, abs_tol=1e-2)
    assert isclose(gender_biased_w2v_small.calc_indirect_bias('softball',
                                                              'homemaker'),
                   0 if all_zero else 0.38, abs_tol=1e-2)

    assert isclose(gender_biased_w2v_small.calc_indirect_bias('football',
                                                              'footballer'),
                   0 if all_zero else 0.02, abs_tol=1e-2)
    # this word have legit gender direction projection
    if not all_zero:
        # TODO in the article it is 0.31 - why?
        assert isclose(gender_biased_w2v_small.calc_indirect_bias('football',
                                                                  'businessman'),  # pylint: disable=C0301
                       0 if all_zero else 0.17, abs_tol=1e-2)
    assert isclose(gender_biased_w2v_small.calc_indirect_bias('football',
                                                              'pundit'),
                   0 if all_zero else 0.10, abs_tol=1e-2)
    assert isclose(gender_biased_w2v_small.calc_indirect_bias('football',
                                                              'maestro'),
                   0 if all_zero else 0.41, abs_tol=1e-2)
    assert isclose(gender_biased_w2v_small.calc_indirect_bias('football',
                                                              'cleric'),
                   0 if all_zero else 0.02, abs_tol=1e-2)


def test_generate_closest_words_indirect_bias(gender_biased_w2v_small):
    """Test generate_closest_words_indirect_bias in GenderBiasWE."""
    result = {'indirect_bias': {('football', 'businessman'): 0.17,
                                ('football', 'cleric'): 0.02,
                                ('football', 'footballer'): 0.02,
                                ('football', 'maestro'): 0.42,
                                ('football', 'pundit'): 0.1,
                                ('softball', 'bookkeeper'): 0.2,
                                ('softball', 'paralegal'): 0.37,
                                ('softball', 'receptionist'): 0.67,
                                ('softball', 'registered_nurse'): 0.29,
                                ('softball', 'waitress'): 0.32},
              'projection': {('football', 'businessman'): -0.2,
                             ('football', 'cleric'): -0.17,
                             ('football', 'footballer'): -0.34,
                             ('football', 'maestro'): -0.18,
                             ('football', 'pundit'): -0.19,
                             ('softball', 'bookkeeper'): 0.18,
                             ('softball', 'paralegal'): 0.14,
                             ('softball', 'receptionist'): 0.16,
                             ('softball', 'registered_nurse'): 0.16,
                             ('softball', 'waitress'): 0.15}}

    indirect_bias_df = (gender_biased_w2v_small
                        .generate_closest_words_indirect_bias('softball',
                                                              'football'))
    assert (indirect_bias_df
            .round(2)
            .to_dict()) == result


def check_all_vectors_unit_length(bias_we):
    for word in bias_we.model.vocab:
        vector = bias_we[word]
        norm = (vector ** 2).sum()
        np.testing.assert_allclose(norm, 1, atol=ATOL)


def test_neutralize(gender_biased_w2v_small, is_preforming=True):
    """Test _neutralize method in GenderBiasWE."""
    neutral_words = gender_biased_w2v_small._data['neutral_words']

    if is_preforming:
        gender_biased_w2v_small._neutralize(neutral_words)

    direction_projections = [project_vector(gender_biased_w2v_small[word],
                                            gender_biased_w2v_small.direction)
                             for word in neutral_words]

    np.testing.assert_allclose(direction_projections, 0, atol=ATOL)

    np.testing.assert_allclose(gender_biased_w2v_small.calc_direct_bias(), 0,
                               atol=ATOL)

    check_all_vectors_unit_length(gender_biased_w2v_small)
    test_calc_indirect_bias(gender_biased_w2v_small, all_zero=True)


def test_equalize(gender_biased_w2v_small, is_preforming=True):
    """Test _equalize method in GenderBiasWE."""
    # pylint: disable=line-too-long
    equality_sets = {tuple(w) for w in gender_biased_w2v_small._data['equalize_pairs']}
    equality_sets |= {tuple(w) for w in gender_biased_w2v_small._data['definitional_pairs']}
    equality_sets = gender_biased_w2v_small._generate_pair_candidates(equality_sets)

    if is_preforming:
        gender_biased_w2v_small._equalize(equality_sets)

    for equality_set in equality_sets:
        projection_vectors = []
        rejection_vectors = []

        for equality_word in equality_set:
            vector = gender_biased_w2v_small[equality_word]

            np.testing.assert_allclose(np.linalg.norm(vector), 1, atol=ATOL)

            (projection_vector,
             rejection_vector) = project_reject_vector(vector,
                                                       gender_biased_w2v_small.direction)
            projection_vectors.append(projection_vector)
            rejection_vectors.append(rejection_vector)

        # <e1, d> == -<e2, d>
        # assuming equality sets of size 2
        assert len(projection_vectors) == 2
        np.testing.assert_allclose(projection_vectors[0] @ gender_biased_w2v_small.direction,
                                   -projection_vectors[1] @ gender_biased_w2v_small.direction,
                                   atol=ATOL)

        # all rejection part is equal for all the vectors
        for rejection_vector in rejection_vectors[1:]:
            np.testing.assert_allclose(rejection_vectors[0],
                                       rejection_vector,
                                       atol=ATOL)

    check_all_vectors_unit_length(gender_biased_w2v_small)


def test_hard_debias_inplace(gender_biased_w2v_small, is_preforming=True):
    """Test hard_debias method in GenderBiasWE."""
    # pylint: disable=C0301
    if is_preforming:
        test_calc_direct_bias(gender_biased_w2v_small)
        gender_biased_w2v_small.debias(method='hard')

    test_neutralize(gender_biased_w2v_small, is_preforming=False)
    test_equalize(gender_biased_w2v_small, is_preforming=False)

    equality_sets = gender_biased_w2v_small._data['definitional_pairs']

    np.random.seed(RANDOM_STATE)
    neutral_words = np.random.choice(gender_biased_w2v_small._data['neutral_words'],
                                     N_RANDOM_NEUTRAL_WORDS_DEBIAS_TO_TEST,
                                     replace=False)

    # for every neutal word w: <e1, w> == <e2, w> AND ||e1 - w|| == ||e2 - w||
    for neutral_word in neutral_words:
        for equality_word1, equality_word2 in equality_sets:

            we1 = gender_biased_w2v_small[neutral_word] @ gender_biased_w2v_small[equality_word1]
            we2 = gender_biased_w2v_small[neutral_word] @ gender_biased_w2v_small[equality_word2]
            np.testing.assert_allclose(we1, we2, atol=ATOL)

            we1_distance = np.linalg.norm(gender_biased_w2v_small[neutral_word]
                                          - gender_biased_w2v_small[equality_word1])
            we2_distance = np.linalg.norm(gender_biased_w2v_small[neutral_word]
                                          - gender_biased_w2v_small[equality_word2])

            np.testing.assert_allclose(we1_distance, we2_distance, atol=ATOL)


def test_hard_debias_not_inplace(gender_biased_w2v_small):
    test_calc_direct_bias(gender_biased_w2v_small)

    gender_debiased_we = gender_biased_w2v_small.debias(method='hard',
                                                        inplace=False)

    test_calc_direct_bias(gender_biased_w2v_small)
    test_hard_debias_inplace(gender_debiased_we, is_preforming=False)


def test_copy(gender_biased_w2v_small):
    gender_biased_w2v_small_copy = copy.copy(gender_biased_w2v_small)
    assert (gender_biased_w2v_small.direction
            is not gender_biased_w2v_small_copy.direction)
    assert gender_biased_w2v_small.model is gender_biased_w2v_small_copy.model


def test_deepcopy(gender_biased_w2v_small):
    gender_biased_w2v_small_copy = copy.deepcopy(gender_biased_w2v_small)
    assert (gender_biased_w2v_small.direction
            is not gender_biased_w2v_small_copy.direction)
    assert (gender_biased_w2v_small.model
            is not gender_biased_w2v_small_copy.model)


def test_evaluate_word_embedding(gender_biased_w2v_small):
    """Test evaluate_word_embedding method in GenderBiasWE."""
    # pylint: disable=C0301
    (word_pairs_evaluation,
     word_analogies_evaluation) = gender_biased_w2v_small.evaluate_word_embedding()

    assert (word_pairs_evaluation.to_dict()
            == {'pearson_r': {'WS353': 0.645, 'RG65': 0.576, 'RW': 0.611, 'Mturk': 0.65, 'MEN': 0.766, 'SimLex999': 0.456, 'TR9856': 0.666},
                'pearson_pvalue': {'WS353': 0.0, 'RG65': 0.232, 'RW': 0.0, 'Mturk': 0.0, 'MEN': 0.0, 'SimLex999': 0.0, 'TR9856': 0.0},
                'spearman_r': {'WS353': 0.688, 'RG65': 0.493, 'RW': 0.655, 'Mturk': 0.674, 'MEN': 0.782, 'SimLex999': 0.444, 'TR9856': 0.676},
                'spearman_pvalue': {'WS353': 0.0, 'RG65': 0.321, 'RW': 0.0, 'Mturk': 0.0, 'MEN': 0.0, 'SimLex999': 0.0, 'TR9856': 0.0},
                'ratio_unkonwn_words': {'WS353': 9.915, 'RG65': 14.286, 'RW': 77.384, 'Mturk': 1.558, 'MEN': 15.148, 'SimLex999': 1.702, 'TR9856': 89.722}})

    assert (word_analogies_evaluation.to_dict()
            == {'score': {'MSR-syntax': 0.75, 'Google': 0.729}})


def test_generate_analogies(gender_biased_w2v_small):
    """Test generate_analogies method in GenderBiasWE.

    Based on:
    https://github.com/tolga-b/debiaswe/blob/master/tutorial_example1.ipynb
    """

    analogies_df = (gender_biased_w2v_small
                    .generate_analogies(500, unrestricted=False))
    analogies_df = analogies_df[['she', 'he']]
    assert_deep_almost_equal(analogies_df.values, TOLGA_GENDER_ANALOGIES)


# TODO deeper testing, this is barely checking it runs
# TODO not all full_specific_words are lower case - why? maybe just names?
# TODO maybe it was trained on the whole w2v?
def test_learn_full_specific_words(gender_biased_w2v_small):
    (full_specific_words,
     clf, X, y) = gender_biased_w2v_small.learn_full_specific_words(debug=True)
    full_specific_words.sort()
    assert (set(gender_biased_w2v_small._data['specific_seed'])
            .issubset(full_specific_words))


def test_calc_all_weat(w2v_small):
    calc_all_weat(w2v_small, filter_by='model', with_original_finding=True,
                  with_pvalue=True, pvalue_kwargs={'method': 'approximate'})


def test_calc_all_weat_index(w2v_small):
    all_weat = calc_all_weat(w2v_small, filter_by='model',
                             with_original_finding=True,
                             with_pvalue=True,
                             pvalue_kwargs={'method': 'approximate'})

    for index in range(len(WEAT_DATA)):
        single_weat = calc_all_weat(w2v_small, weat_data=index,
                                    filter_by='model',
                                    with_original_finding=True,
                                    with_pvalue=True,
                                    pvalue_kwargs={'method':
                                                   'approximate'})

        assert_deep_almost_equal(single_weat.iloc[0].to_dict(),
                                 all_weat.iloc[index].to_dict(),
                                 atol=0.01)


def test_calc_all_weat_indices(w2v_small):
    all_weat = calc_all_weat(w2v_small, filter_by='model',
                             with_original_finding=True,
                             with_pvalue=True,
                             pvalue_kwargs={'method': 'approximate'})

    for index in range(1, len(WEAT_DATA)):
        indices = tuple(range(index))
        singles_weat = calc_all_weat(w2v_small, weat_data=indices,
                                     filter_by='model',
                                     with_original_finding=True,
                                     with_pvalue=True,
                                     pvalue_kwargs={'method':
                                                    'approximate'})

        assert_deep_almost_equal(singles_weat.to_dict(),
                                 all_weat.iloc[:index].to_dict(),
                                 atol=0.01)


def test_calc_all_weat_defaults(w2v_small):
    weat_5_results_default = (calc_all_weat(w2v_small, (5,))
                              .iloc[0].to_dict())

    weat_5_results = (calc_all_weat(w2v_small, (5,),
                                    filter_by='model',
                                    with_pvalue=True,
                                    pvalue_kwargs={'method': 'exact'})
                      .iloc[0].to_dict())

    assert_deep_almost_equal(weat_5_results_default, weat_5_results)


def test_calc_weat_pleasant_attribute(w2v_small):
    # pylint: disable=line-too-long

    pvalue_kwargs = {'method': 'approximate'}

    result_v1 = calc_weat_pleasant_unpleasant_attribute(w2v_small,
                                                        WEAT_DATA[1]['first_target'],
                                                        WEAT_DATA[1]['second_target'],
                                                        pvalue_kwargs=pvalue_kwargs)
    result_v1['p'] = round(result_v1['p'], 4)
    result_v1['d'] = round(result_v1['d'], 4)
    result_v1['s'] = round(result_v1['s'], 4)

    result_v2 = (calc_all_weat(w2v_small, (1,),
                               pvalue_kwargs=pvalue_kwargs)
                 .iloc[0]
                 .to_dict())

    assert_deep_almost_equal(result_v1, result_v2, atol=0.01)


def test_most_similar(w2v_small):
    POSITIVE, NEGATIVE = ('doctor', 'she'), ('he',)

    responsibly_results = most_similar(w2v_small, POSITIVE, NEGATIVE,
                                       topn=10)
    gensim_results = w2v_small.most_similar(POSITIVE, NEGATIVE,
                                            topn=9)

    assert responsibly_results[0][0] == 'doctor'

    assert_deep_almost_equal(responsibly_results[1:], gensim_results,
                             atol=0.01)


def test_compute_association(gender_biased_w2v_small):
    """
    Test compute_association method in GenderBiasWE.
    """
    (r, pvalue), _ = gender_biased_w2v_small.compute_factual_association()

    assert isclose(r, 0.7070401592764508, abs_tol=ATOL)
    assert isclose(pvalue, 1.4324502214459908e-06, abs_tol=ATOL)



import json

import numpy as np
import pandas as pd
from pkg_resources import resource_filename, resource_stream

from responsibly.dataset.core import Dataset


__all__ = ['GermanDataset']

GERMAN_PATH = resource_filename(__name__,
                                'german.data')

VALUES_MAPS = json.loads(resource_stream(__name__,
                                         'values_maps.json')
                         .read()
                         .decode())

COLUMN_NAMES = ['status', 'duration', 'credit_history', 'purpose',
                'credit_amount', 'savings', 'present_employment',
                'installment_rate', 'status_sex', 'other_debtors',
                'present_residence_since', 'property', 'age',
                'installment_plans', 'housing',
                'number_of_existing_credits', 'job',
                'number_of_people_liable_for', 'telephone',
                'foreign_worker', 'credit']


class GermanDataset(Dataset):
    """German Credit Dataset.

    See :class:`~responsibly.dataset.Dataset` for a description of
    the arguments and attributes.

    References:
        - https://archive.ics.uci.edu/ml/datasets/statlog+(german+credit+data)
        - Kamiran, F., & Calders, T. (2009, February).
          Classifying without discriminating.
          In 2009 2nd International Conference on Computer, Control
          and Communication (pp. 1-6). IEEE.
          http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.182.6067&rep=rep1&type=pdf


    Extra
        This dataset requires use of a cost matrix (see below)

        ::

               1 2
               ----
            1 | 0 1
              |----
            2 | 5 0

        (1 = Good, 2 = Bad)

        The rows represent the actual classification
        and the columns the predicted classification.
        It is worse to class a customer as good when they are bad (5),
        than it is to class a customer as bad when they are good (1).

    """

    def __init__(self):
        super().__init__(target='credit',
                         sensitive_attributes=['age_factor'])
        self.cost_matrix = [[0, 1], [5, 0]]

    def _load_data(self):
        return pd.read_csv(GERMAN_PATH, sep=' ', names=COLUMN_NAMES,
                           header=None, index_col=False)

    def _preprocess(self):
        """Perform the same preprocessing as the dataset doc file."""
        self.df['credit'] = self.df['credit'].astype(str)

        for col, translation in VALUES_MAPS.items():
            self.df[col] = self.df[col].map(translation)

        new_column_names = COLUMN_NAMES[:]

        self.df['status'], self.df['sex'] = (self.df['status_sex']
                                             .str
                                             .split(' : ')
                                             .str)
        self.df = self.df.drop('status_sex', axis=1)

        status_sex_index = new_column_names.index('status_sex')
        new_column_names[status_sex_index:status_sex_index + 1] = \
            ['status', 'sex']

        self.df['age_factor'] = pd.cut(self.df['age'],
                                       [19, 25, 76],
                                       right=False)
        age_factor_index = new_column_names.index('age') + 1
        new_column_names.insert(age_factor_index, 'age_factor')

        self.df = self.df[new_column_names]

    def _validate(self):
        # pylint: disable=line-too-long
        super()._validate()

        assert len(self.df) == 1000, 'the number of rows should be 1000,'\
                                     ' but it is {}.'.format(len(self.df))
        assert len(self.df.columns) == 23, 'the number of columns should be 23,'\
                                           ' but it is {}.'.format(len(self.df.columns))
        assert not self.df.isnull().any().any(), 'there are null values.'
        assert self.df['age_factor'].nunique() == 2,\
            'age_factor should have only 2 unique values,'\
            ' but it is{}'.format(self.df['age_factor'].nunique())



"""Unit test module for responsibly.fairness

FICO results are based on running the code from:
https://github.com/fairmlbook/fairmlbook.github.io
"""
# pylint: disable=redefined-outer-name,line-too-long

import pytest

from responsibly.dataset import COMPASDataset, build_FICO_dataset
from responsibly.fairness.interventions import threshold
from responsibly.fairness.metrics import (
    independence_binary, plot_roc_by_attr, separation_binary,
    sufficiency_binary,
)
from responsibly.fairness.metrics.binary import compare_privileged
from responsibly.tests.utils import assert_deep_almost_equal


FICO_TOL = {'atol': 1e-5, 'rtol': 5e-3}

COST_MATRIX = [[0, - 5 / 6], [0, 1 / 6]]

FICO_SINGLE = (42.5,
               {'Asian': (0.159210978241477, 0.7982257394726964),
                'Black': (0.03562475202516957, 0.5222536366818645),
                'Hispanic': (0.0811160539719404, 0.6318581559145164),
                'White': (0.10494970149747886, 0.7974084986440749)},
               -11888.109525713331 / 174047)

# In the FairMLBook repo the cutoffs are
# {'Asian': 38.5, 'Black': 49.0, 'Hispanic': 47.0, 'White': 41.0}
# Cost per group from there:
# max_profit Asian 0.0027352900000000044
# max_profit Black 1.0889999999999889e-05
# max_profit Hispanic 9.025000000000055e-05
# max_profit White 0.00021315999999999818
# Cost is NOT taken from FairMLBook repo
FICO_MIN_COST = ({'Asian': 38.5,
                  'Black': 49.0,
                  'Hispanic': 47.0,
                  'White': 41.0},
                 {'Asian': (0.18688340039863297, 0.8409289340716826),
                  'Black': (0.020748274408693028, 0.4148991737082274),
                  'Hispanic': (0.05865353927132363, 0.5589875936027977),
                  'White': (0.11484787793720092, 0.8138995270074796)},
                 -0.06876633661556937)


# In the FairMLBook repo the cutoffs are
# {'Asian': 53.5, 'Black': 18.5, 'Hispanic': 32.0, 'White': 52.5}
# and the acceptance rate is 0.5189755206777411
# It might be that there is some issue with _ternary_search_float
# The cost is slightly worse, so we probably don't find the optimal
# solution, but very close to it
# From the repo this is the cost: -8183.343726279998 / 174047
FICO_INDEPENDENCE = ({'Asian': 53.5,
                      'Black': 18.0,
                      'Hispanic': 31.5,
                      'White': 52.0},
                     {'Asian': (0.08776430097924894, 0.6420627855534083),
                      'Black': (0.3445144172941361, 0.908131231733616),
                      'Hispanic': (0.1739271675548935, 0.8061475266133501),
                      'White': (0.060298201269177995, 0.6844160268999959)},
                     -0.046725728876261405,
                     0.532796101034295)

# All the values are NOT taken from FairMLBook repo
FICO_FNR = ({'Asian': 45.5, 'Black': 29.5, 'Hispanic': 35.0, 'White': 46},
            {'Asian': (0.13830821486672065, 0.7597270937393894),
             'Black': (0.11069747158860148, 0.753996048320934),
             'Hispanic': (0.13779115742038225, 0.7528248235528201),
             'White': (0.0845338139109347, 0.7565831625671293)},
            -0.06352608487155771,
            0.2402515268107051)

FICO_SEPARATION = ({},
                   {'': (0.11130392607695799, 0.7032306899595236)},
                   # slightly different from the fairmlbook repository:
                   -9448.09030585854 / 174047)

FICO_THRESHOLD_DATA = {'single': FICO_SINGLE,
                       'min_cost': FICO_MIN_COST,
                       'independence': FICO_INDEPENDENCE,
                       'fnr': FICO_FNR,
                       'separation': FICO_SEPARATION}


@pytest.fixture
def compas_ds():
    ds = COMPASDataset()
    df = ds.df
    df = df[df['race'].isin(['African-American',
                             'Caucasian'])]
    ds.df = df
    return ds


def test_independence_binary(compas_ds):
    indp, _ = independence_binary((compas_ds.df['y_pred']),
                                  compas_ds.df['race'])
    assert (indp['African-American']['acceptance_rate']
            == pytest.approx(0.576, abs=0.001))
    assert (indp['Caucasian']['acceptance_rate']
            == pytest.approx(0.33, abs=0.001))


def test_sufficiency_binary(compas_ds):
    suf, _ = sufficiency_binary(compas_ds.df['two_year_recid'],
                                compas_ds.df['y_pred'],
                                compas_ds.df['race'])
    assert (suf['African-American']['ppv']
            == pytest.approx(0.649, abs=0.001))
    assert (suf['Caucasian']['ppv']
            == pytest.approx(0.594, abs=0.001))
    assert (suf['African-American']['npv']
            == pytest.approx(0.648, abs=0.001))
    assert (suf['Caucasian']['npv']
            == pytest.approx(0.710, abs=0.001))


def test_separation_binary(compas_ds):
    sep, _ = separation_binary(compas_ds.df['two_year_recid'],
                               compas_ds.df['y_pred'],
                               compas_ds.df['race'])
    assert (sep['African-American']['tpr']
            == pytest.approx(0.715, abs=0.001))
    assert (sep['Caucasian']['tpr']
            == pytest.approx(0.503, abs=0.001))
    assert (sep['African-American']['fpr']
            == pytest.approx(0.423, abs=0.001))
    assert (sep['Caucasian']['fpr']
            == pytest.approx(0.220, abs=0.001))


def test_compare_privileged():
    assert (compare_privileged({'A': {'x': 1, 'y': 40},
                                'B': {'x': 4, 'y': 5}},
                               'B')
            == {'metrics': {'x': {'diff': -3, 'ratio': 0.25},
                            'y': {'diff': 35, 'ratio': 8.0}},
                'x_sens_privileged': 'B',
                'x_sens_unprivileged': 'A'})


@pytest.fixture
def fico():
    return build_FICO_dataset()


def test_single_threshold(fico):
    assert_deep_almost_equal(FICO_SINGLE,
                             threshold.find_single_threshold(fico['rocs'],
                                                             fico['base_rates'],
                                                             fico['proportions'],
                                                             COST_MATRIX),
                             **FICO_TOL)


# https://github.com/fairmlbook/fairmlbook.github.io/blob/b1dfc9756ee8d2adf502c1b0f7265f10c5b1033c/code/creditscore/criteria.py#L122
# in the github repo of fairmlbook, for max profit case
# the threshold for Hispanic is `47.0`,
# and not `46.5` as we get with our method.
# maybe it is because we are using trinary search for finding the minimum
# on the cost function, and in the fairmlbook there are performing argmin
# all over the cost function values.
def test_min_cost_threshold(fico):
    assert_deep_almost_equal(FICO_MIN_COST,
                             threshold.find_min_cost_thresholds(fico['rocs'],
                                                                fico['base_rates'],
                                                                fico['proportions'],
                                                                COST_MATRIX),
                             **FICO_TOL)


def test_independence_thresholds(fico):
    assert_deep_almost_equal(FICO_INDEPENDENCE,
                             threshold.find_independence_thresholds(fico['rocs'],
                                                                    fico['base_rates'],
                                                                    fico['proportions'],
                                                                    COST_MATRIX),
                             **FICO_TOL)


def test_fnr_thresholds(fico):
    assert_deep_almost_equal(FICO_FNR,
                             threshold.find_fnr_thresholds(fico['rocs'],
                                                           fico['base_rates'],
                                                           fico['proportions'],
                                                           COST_MATRIX),
                             **FICO_TOL)


def test_separation_thresholds(fico):
    assert_deep_almost_equal(FICO_SEPARATION,
                             threshold.find_separation_thresholds(fico['rocs'],
                                                                  fico['base_rate'],
                                                                  COST_MATRIX),
                             **FICO_TOL)


def test_thresholds(fico):
    threshold_data = threshold.find_thresholds(fico['rocs'],
                                               fico['proportions'],
                                               fico['base_rate'],
                                               fico['base_rates'],
                                               COST_MATRIX)

    assert_deep_almost_equal(FICO_THRESHOLD_DATA, threshold_data,
                             **FICO_TOL)


def test_plot_roc_by_attr_thresholds_exception(compas_ds):
    df_missing_score = compas_ds.df[(((compas_ds.df['race'] == 'Caucasian')
                                      & (compas_ds.df['decile_score'] != 5))
                                     | (compas_ds.df['race'] == 'African-American'))]

    with pytest.raises(NotImplementedError):
        plot_roc_by_attr(df_missing_score['two_year_recid'],
                         df_missing_score['decile_score'],
                         df_missing_score['race'])



"""
Demographic Classification Fairness Criteria.

The objectives of the demographic classification fairness criteria
is to measure unfairness towards sensitive attribute valuse.

The metrics have the same interface and behavior as the ones in
:mod:`sklearn.metrics`
(e.g., using ``y_true``, ``y_pred`` and ``y_score``).

One should keep in mind that the criteria are intended
to *measure unfairness, rather than to prove fairness*, as it stated in
the paper `Equality of opportunity in supervised learning <https://arxiv.org/abs/1610.02413>`_
by Hardt et al. (2016):

    ... satisfying [the demographic criteria] should not be
    considered a conclusive proof of fairness.
    Similarly, violations of our condition are not meant
    to be a proof of unfairness.
    Rather we envision our framework as providing a reasonable way
    of discovering and measuring potential concerns that require
    further scrutiny. We believe that resolving fairness concerns is
    ultimately impossible without substantial domain-specific
    investigation.

The output of binary classifiers can come in two forms, either giving
a binary outcome prediction for input or producing
a real number score, which the common one is the probability
for the positive or negative label
(such as the method ``proba`` of an ``Estimator`` in ``sklearn``).
Therefore, the criteria come in two flavors, one for **binary** output,
and the second for **score** output.

The fundamental concept for defining the fairness criteria
is `conditional independence <https://en.wikipedia.org/wiki/Conditional_independence>`_.
Using *Machine Learning and Fairness* book's notions:

- ``A`` - Sensitive attribute
- ``Y`` - Binary ground truth (correct) target
- ``R`` - Estimated binary targets or score as returned by a classifier

There are three demographic fairness criteria for classification:

1. Independence - R⊥A

2. Separation - R⊥A∣Y

3. Sufficiency - Y⊥A∣R

"""


from responsibly.fairness.metrics.binary import (
    independence_binary, report_binary, separation_binary, sufficiency_binary,
)
from responsibly.fairness.metrics.score import (
    independence_score, roc_auc_score_by_attr, roc_curve_by_attr,
    separation_score, sufficiency_score,
)
from responsibly.fairness.metrics.visualization import (
    distplot_by, plot_roc_by_attr, plot_roc_curves,
)



TOLGA_GENDER_ANALOGIES = [('she', 'he'),
                          ('herself', 'himself'),
                          ('her', 'his'),
                          ('woman', 'man'),
                          ('daughter', 'son'),
                          ('businesswoman', 'businessman'),
                          ('girl', 'boy'),
                          ('actress', 'actor'),
                          ('chairwoman', 'chairman'),
                          ('heroine', 'hero'),
                          ('mother', 'father'),
                          ('spokeswoman', 'spokesman'),
                          ('sister', 'brother'),
                          ('girls', 'boys'),
                          ('sisters', 'brothers'),
                          ('queen', 'king'),
                          ('niece', 'nephew'),
                          ('councilwoman', 'councilman'),
                          ('motherhood', 'fatherhood'),
                          ('women', 'men'),
                          ('petite', 'lanky'),
                          ('ovarian_cancer', 'prostate_cancer'),
                          ('Anne', 'John'),
                          ('schoolgirl', 'schoolboy'),
                          ('granddaughter', 'grandson'),
                          ('aunt', 'uncle'),
                          ('matriarch', 'patriarch'),
                          ('twin_sister', 'twin_brother'),
                          ('mom', 'dad'),
                          ('lesbian', 'gay'),
                          ('husband', 'younger_brother'),
                          ('gal', 'dude'),
                          ('lady', 'gentleman'),
                          ('sorority', 'fraternity'),
                          ('mothers', 'fathers'),
                          ('grandmother', 'grandfather'),
                          ('blouse', 'shirt'),
                          ('soprano', 'baritone'),
                          ('queens', 'kings'),
                          ('Jill', 'Greg'),
                          ('daughters', 'sons'),
                          ('grandma', 'grandpa'),
                          ('volleyball', 'football'),
                          ('diva', 'superstar'),
                          ('mommy', 'kid'),
                          ('Sarah', 'Matthew'),
                          ('hairdresser', 'barber'),
                          ('softball', 'baseball'),
                          ('goddess', 'god'),
                          ('Aisha', 'Jamal'),
                          ('waitress', 'waiter'),
                          ('princess', 'prince'),
                          ('filly', 'colt'),
                          ('mare', 'gelding'),
                          ('ladies', 'gentlemen'),
                          ('childhood', 'boyhood'),
                          ('interior_designer', 'architect'),
                          ('nun', 'priest'),
                          ('wig', 'beard'),
                          ('granddaughters', 'grandsons'),
                          ('girlfriends', 'buddies'),
                          ('gals', 'dudes'),
                          ('aunts', 'uncles'),
                          ('congresswoman', 'congressman'),
                          ('feminism', 'conservatism'),
                          ('bitch', 'bastard'),
                          ('hers', 'yours'),
                          ('bra', 'pants'),
                          ('moms', 'dads'),
                          ('nurse', 'surgeon'),
                          ('heiress', 'magnate'),
                          ('feminine', 'manly'),
                          ('glamorous', 'flashy'),
                          ('actresses', 'actors'),
                          ('registered_nurse', 'physician'),
                          ('cupcakes', 'pizzas'),
                          ('blond', 'burly'),
                          ('babe', 'fella'),
                          ('mums', 'blokes'),
                          ('gorgeous', 'magnificent'),
                          ('compatriot', 'countryman'),
                          ('fabulous', 'terrific'),
                          ('breast', 'prostate'),
                          ('starlet', 'youngster'),
                          ('Laurie', 'Brett'),
                          ('kids', 'guys'),
                          ('sewing', 'carpentry'),
                          ('kinda', 'guy'),
                          ('headscarf', 'turban'),
                          ('siblings', 'elder_brother'),
                          ('charming', 'affable'),
                          ('sassy', 'snappy'),
                          ('cosmetics', 'pharmaceuticals'),
                          ('estrogen', 'testosterone'),
                          ('handbag', 'briefcase'),
                          ('housewife', 'shopkeeper'),
                          ('fillies', 'colts'),
                          ('nieces', 'nephews'),
                          ('whore', 'coward'),
                          ('boyfriend', 'pal'),
                          ('salon', 'barbershop'),
                          ('Latonya', 'Leroy'),
                          ('vagina', 'penis'),
                          ('breast_cancer', 'lymphoma'),
                          ('vocalist', 'guitarist'),
                          ('me', 'him'),
                          ('children', 'youngsters'),
                          ('adorable', 'goofy'),
                          ('giggling', 'grinning'),
                          ('cheerful', 'jovial'),
                          ('lovely', 'brilliant'),
                          ('giggle', 'chuckle'),
                          ('bras', 'trousers'),
                          ('wedding_dress', 'tuxedo'),
                          ('singer', 'frontman'),
                          ('netball', 'rugby'),
                          ('rebounder', 'playmaker'),
                          ('vocalists', 'trumpeter'),
                          ('nude', 'shirtless'),
                          ('beautiful', 'majestic'),
                          ('feisty', 'mild_mannered'),
                          ('feminists', 'socialists'),
                          ('nanny', 'chauffeur'),
                          ('females', 'males'),
                          ('pediatrician', 'orthopedic_surgeon'),
                          ('teenage_girls', 'youths'),
                          ('pink', 'red'),
                          ('convent', 'monastery'),
                          ('midwife', 'doctor'),
                          ('feminist', 'liberal'),
                          ('gown', 'blazer'),
                          ('blonde', 'blond'),
                          ('stepdaughter', 'stepson'),
                          ('wonderful', 'great'),
                          ('breasts', 'genitals'),
                          ('luscious', 'crisp'),
                          ('judgmental', 'arrogant'),
                          ('skirts', 'shorts'),
                          ('middle_aged', 'bearded'),
                          ('spokespeople', 'spokesmen'),
                          ('female', 'male'),
                          ('beauty', 'grandeur'),
                          ('salesperson', 'salesman'),
                          ('witch', 'demon'),
                          ('male_counterparts', 'counterparts'),
                          ('violinist', 'virtuoso'),
                          ('practicality', 'durability'),
                          ('boobs', 'ass'),
                          ('dolls', 'replicas'),
                          ('husbands', 'wives'),
                          ('ponytail', 'mustache'),
                          ('sexism', 'racism'),
                          ('mammogram', 'colonoscopy'),
                          ('sweater', 'jersey'),
                          ('hysterical', 'comical'),
                          ('uterus', 'intestine'),
                          ('rehearse', 'improvise'),
                          ('classmates', 'teammates'),
                          ('stroller', 'bicycle'),
                          ('presenter', 'broadcaster'),
                          ('friends', 'buddy'),
                          ('dresses', 'shirts'),
                          ('eating_disorders', 'alcoholism'),
                          ('kindness', 'humility'),
                          ('cute', 'clever'),
                          ('sobbed', 'grinned'),
                          ('baking', 'roasting'),
                          ('sweaters', 'jerseys'),
                          ('crafting', 'drafting'),
                          ('nuns', 'monk'),
                          ('lesbians', 'homosexual'),
                          ('guidance_counselor', 'headmaster'),
                          ('terrifying', 'fearsome'),
                          ('libero', 'midfielders'),
                          ('classmate', 'teammate'),
                          ('underclassmen', 'players'),
                          ('pianist', 'maestro'),
                          ('boyfriends', 'friends'),
                          ('teenager', 'lad'),
                          ('giggles', 'chuckles'),
                          ('gender', 'racial'),
                          ('hair_salon', 'pizzeria'),
                          ('pitcher', 'starter'),
                          ('lovers', 'aficionados'),
                          ('lingerie', 'menswear'),
                          ('kittens', 'neutered'),
                          ('latex', 'rubber'),
                          ('dress', 'garb'),
                          ('alluring', 'intriguing'),
                          ('buns', 'sausages'),
                          ('cosmetic_surgery', 'surgery'),
                          ('foal', 'stallion'),
                          ('sexy', 'nerdy'),
                          ('delightful', 'superb'),
                          ('cubs', 'lions'),
                          ('self_esteem', 'morale'),
                          ('rower', 'skipper'),
                          ('bingo', 'gambling'),
                          ('figure_skating', 'hockey'),
                          ('stepmother', 'eldest_son'),
                          ('swimwear', 'sportswear'),
                          ('teen', 'youth'),
                          ('gymnastics', 'weightlifting'),
                          ('tissues', 'cartilage'),
                          ('enchanting', 'splendid'),
                          ('hooker', 'winger'),
                          ('graphic_designer', 'carpenter'),
                          ('layups', 'downfield'),
                          ('sophomores', 'rookies'),
                          ('singers', 'guitarists'),
                          ('backcourt', 'playmakers'),
                          ('hugs', 'shook_hands'),
                          ('sexist', 'racist'),
                          ('cheesecake', 'pizza'),
                          ('sultry', 'mellow'),
                          ('silicone', 'polymer'),
                          ('gymnasts', 'athletes'),
                          ('ribbons', 'flags'),
                          ('captivating', 'electrifying'),
                          ('gymnast', 'athlete'),
                          ('cried', 'chuckled'),
                          ('ethereal', 'brooding'),
                          ('cats', 'pigeons'),
                          ('implants', 'stents'),
                          ('cougar', 'wolves'),
                          ('fiance', 'cousin'),
                          ('cheery', 'amiable'),
                          ('servicemen', 'veterans'),
                          ('hair', 'facial_hair'),
                          ('sophomore', 'pounder'),
                          ('midwives', 'doctors'),
                          ('chairperson', 'managing_director'),
                          ('meter_hurdles', 'yard_dash'),
                          ('scream', 'boo'),
                          ('glam', 'retro'),
                          ('terrified', 'mad'),
                          ('shrill', 'louder'),
                          ('lupus', 'multiple_myeloma'),
                          ('uptight', 'cocky'),
                          ('hubby', 'pals'),
                          ('exclaimed', 'quipped'),
                          ('choreography', 'footwork'),
                          ('veil', 'cloak'),
                          ('panties', 'socks'),
                          ('gowns', 'robes'),
                          ('designers', 'architects'),
                          ('foul_trouble', 'defensively'),
                          ('amazing', 'unbelievable'),
                          ('artistry', 'genius'),
                          ('sensual', 'moody'),
                          ('campers', 'camp'),
                          ('Allison', 'Todd'),
                          ('freaking', 'heck'),
                          ('intuition', 'gut_feeling'),
                          ('eating_disorder', 'addiction'),
                          ('frontcourt', 'swingman'),
                          ('hostess', 'bartender'),
                          ('child_endangerment', 'aggravated_assault'),
                          ('masculine', 'macho'),
                          ('novelist', 'philosopher'),
                          ('putback', 'yarder'),
                          ('ex_boyfriend', 'ex_girlfriend'),
                          ('incredibly', 'obviously'),
                          ('beautify', 'rehabilitate'),
                          ('designed', 'engineered'),
                          ('mares', 'thoroughbred'),
                          ('satin', 'leather'),
                          ('hurler', 'lefthander'),
                          ('bun', 'burger'),
                          ('heartbreaking', 'humbling'),
                          ('romantic_comedy', 'flick'),
                          ('fourteen', 'eleven'),
                          ('flight_attendant', 'pilots'),
                          ('cupcake', 'donut'),
                          ('glitter', 'confetti'),
                          ('sobbing', 'smiled'),
                          ('unassisted_goal', 'powerplay'),
                          ('memoir', 'autobiography'),
                          ('cigarette', 'cigar'),
                          ('animal_cruelty', 'dogfighting'),
                          ('baby', 'daddy'),
                          ('outfits', 'uniforms'),
                          ('bride', 'groom'),
                          ('sprinter', 'speedster'),
                          ('estranged_husband', 'stepfather'),
                          ('singer_songwriter', 'musician'),
                          ('entrepreneurs', 'businessmen'),
                          ('grad', 'alumnus'),
                          ('lover', 'enthusiast'),
                          ('celebrities', 'superstars'),
                          ('kissing', 'shaking_hands'),
                          ('auditions', 'tryout'),
                          ('backstretch', 'straightaway'),
                          ('pregnancy', 'gestation'),
                          ('librarian', 'curator'),
                          ('anime', 'videogames'),
                          ('heartbroken', 'dejected'),
                          ('cervical_cancer', 'pancreatic_cancer'),
                          ('regionals', 'playoffs'),
                          ('Carrie', 'Laurie'),
                          ('teenage_girl', 'teenager'),
                          ('coworkers', 'colleagues'),
                          ('sergeants', 'lieutenants'),
                          ('paralegal', 'accountant'),
                          ('stressful', 'frustrating'),
                          ('duets', 'saxophonist'),
                          ('plucky', 'hapless'),
                          ('realtor', 'builder'),
                          ('cocktails', 'beers'),
                          ('closets', 'lockers'),
                          ('doubleheader', 'game'),
                          ('thighs', 'backside'),
                          ('wellness', 'fitness'),
                          ('giant_slalom', 'skied'),
                          ('neurotic', 'eccentric'),
                          ('coordinator', 'manager'),
                          ('antiques', 'memorabilia'),
                          ('dancers', 'drummers'),
                          ('layup', 'touchdown_pass'),
                          ('manga', 'comic_books'),
                          ('horrified', 'disappointed'),
                          ('empowering', 'motivating'),
                          ('creepy', 'menacing'),
                          ('insecurities', 'frustrations'),
                          ('middle_blocker', 'leadoff_hitter'),
                          ('adored', 'revered'),
                          ('exercises', 'drills'),
                          ('skirted', 'evaded'),
                          ('jewelry', 'collectibles'),
                          ('choral', 'composer'),
                          ('confided', 'intimated'),
                          ('dermatologist', 'neurosurgeon'),
                          ('attractiveness', 'competitiveness'),
                          ('housekeeper', 'janitor'),
                          ('thyroid', 'inflammation'),
                          ('vocals', 'drummer'),
                          ('scarf', 'jacket'),
                          ('burlesque', 'rock_n_roll'),
                          ('athlete', 'player'),
                          ('chic', 'minimalist'),
                          ('cerebral_palsy', 'spinal_cord_injury'),
                          ('horrid', 'woeful'),
                          ('transgender', 'homophobic'),
                          ('clique', 'inner_circle'),
                          ('pampering', 'luxurious'),
                          ('associate_professor', 'professor_emeritus'),
                          ('sexuality', 'homosexuality'),
                          ('reunite', 'rejoin'),
                          ('caring', 'selfless'),
                          ('freshman', 'redshirt_freshman'),
                          ('plunging', 'slumping'),
                          ('spokesperson', 'statement'),
                          ('prostitution', 'drug_trafficking'),
                          ('characters', 'villain'),
                          ('therapist', 'neurologist'),
                          ('maids', 'servants'),
                          ('mama', 'ya'),
                          ('delightfully', 'brilliantly'),
                          ('boutiques', 'retail_outlets'),
                          ('hormonal', 'metabolism'),
                          ('puppy', 'bulldog'),
                          ('pinch_runner', 'pinch_hitter'),
                          ('behaviors', 'tendencies'),
                          ('appalled', 'displeased'),
                          ('forms', 'form'),
                          ('nonprofit', 'organization'),
                          ('tamoxifen', 'statins'),
                          ('mysterious', 'enigmatic'),
                          ('maternity', 'obstetrics'),
                          ('mentors', 'mentor'),
                          ('worker', 'foreman'),
                          ('anorexia', 'depression'),
                          ('critters', 'beasts'),
                          ('feline', 'rodent'),
                          ('stereotype', 'mentality'),
                          ('starvation', 'famine'),
                          ('fairy', 'magical'),
                          ('gossip', 'rumor_mill'),
                          ('confesses', 'concedes'),
                          ('activists', 'supporters'),
                          ('flower', 'ornamental'),
                          ('contests', 'games'),
                          ('procuring', 'acquiring'),
                          ('silver_medalist', 'compatriot'),
                          ('purse', 'wallet'),
                          ('nurses', 'physicians'),
                          ('sectionals', 'championship'),
                          ('refill', 'recharge'),
                          ('inadequate', 'ineffective'),
                          ('prettiest', 'finest'),
                          ('seatbelt', 'wearing_helmet'),
                          ('rehearsal', 'improvisation'),
                          ('interns', 'fellows'),
                          ('adds', 'added'),
                          ('harmonies', 'guitar'),
                          ('comfy', 'comfortable'),
                          ('menopause', 'heart_disease'),
                          ('scripture', 'disciples'),
                          ('cries', 'chants'),
                          ('experimenting', 'tinkering'),
                          ('underweight', 'underperform'),
                          ('holistic_approach', 'approach'),
                          ('costumes', 'props'),
                          ('curves', 'curve'),
                          ('bartender', 'bouncer'),
                          ('seductive', 'mesmerizing'),
                          ('clarinet', 'trombone'),
                          ('transformative', 'visionary'),
                          ('ok', 'alright'),
                          ('unusual', 'unorthodox'),
                          ('osteoporosis', 'atrial_fibrillation'),
                          ('screams', 'roar'),
                          ('incumbents', 'incumbent'),
                          ('baseman', 'offensive_lineman'),
                          ('strawberry', 'potato'),
                          ('steals', 'tackles'),
                          ('walker', 'crutches'),
                          ('cook', 'grill'),
                          ('workshop', 'meeting'),
                          ('elegance', 'style'),
                          ('emotionally', 'mentally'),
                          ('beloved', 'legendary'),
                          ('breast_milk', 'sperm'),
                          ('minivan', 'pickup'),
                          ('complainants', 'accusers'),
                          ('knitting', 'yarn'),
                          ('artisans', 'craftsmen'),
                          ('strollers', 'scooters'),
                          ('caseload', 'workload'),
                          ('smile', 'grin'),
                          ('helpful', 'useful'),
                          ('caretakers', 'caretaker'),
                          ('blokes', 'bloke'),
                          ('provocative', 'incendiary'),
                          ('ultrasound', 'x_ray'),
                          ('colorful', 'flamboyant'),
                          ('counselor', 'adviser'),
                          ('borrower', 'loan'),
                          ('brightness', 'sharpness'),
                          ('staffing', 'manpower'),
                          ('overwhelmed', 'frustrated'),
                          ('bowlers', 'seamer'),
                          ('bunny', 'monkey'),
                          ('devotion', 'allegiance'),
                          ('advocates', 'critics'),
                          ('brides', 'weddings'),
                          ('cigarettes', 'cigars'),
                          ('dancer', 'entertainer'),
                          ('relationships', 'rapport'),
                          ('attitudes', 'mindset'),
                          ('strives', 'wants'),
                          ('bearish', 'bullish'),
                          ('deadlines', 'deadline'),
                          ('proponents', 'detractors'),
                          ('pitchers', 'ballclub'),
                          ('carriages', 'locomotive'),
                          ('shortstop', 'fullback'),
                          ('exhibit', 'exhibition'),
                          ('credit_crunch', 'slump'),
                          ('watchers', 'pundits'),
                          ('hitter', 'southpaw'),
                          ('scary', 'dangerous'),
                          ('silver_medals', 'championships'),
                          ('expiration_date', 'expires'),
                          ('tenacious', 'hard_nosed'),
                          ('gosh', 'yeah'),
                          ('bright_colors', 'colors'),
                          ('retro', 'throwback'),
                          ('underpass', 'tunnel'),
                          ('thrilled', 'confident'),
                          ('warship', 'destroyer'),
                          ('resignations', 'resignation'),
                          ('reaffirmed', 'reiterated'),
                          ('technical_fouls', 'ejection'),
                          ('sheets', 'sheet'),
                          ('makeover', 'revamp'),
                          ('bookkeeper', 'treasurer'),
                          ('callous', 'foolish'),
                          ('dignified', 'honorable'),
                          ('elegantly', 'superbly'),
                          ('foods', 'foodstuffs'),
                          ('cashier', 'robber'),
                          ('assistant_professor', 'economics_professor'),
                          ('downsize', 'restructure'),
                          ('details', 'specifics'),
                          ('healthful', 'healthy'),
                          ('luminous', 'dazzling'),
                          ('charm', 'swagger'),
                          ('pastry', 'sausage'),
                          ('free_throws', 'recovered_fumble'),
                          ('naysayers', 'doubters'),
                          ('embryo', 'embryonic'),
                          ('stalker', 'pedophile'),
                          ('hormones', 'growth_hormone'),
                          ('irresponsibility', 'recklessness'),
                          ('intangible_assets', 'intangibles'),
                          ('descriptive', 'succinct'),
                          ('screamed', 'shouted'),
                          ('manipulative', 'devious'),
                          ('audio_replay', 'replay')]



# pylint: disable=line-too-long

from responsibly import fairness, we


__project__ = 'responsibly'
__description__ = 'Toolkit for Auditing and Mitigating Bias and Fairness of Machine Learning Systems 🔎🤖🧰'
__url__ = 'https://docs.responsibly.ai'
__version__ = '0.1.3'
__author__ = 'Shlomi Hod'
__author_email__ = 'shlomi.hod@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2018 Shlomi Hod'

VERSION = "{0} v{1}".format(__project__, __version__)



import warnings


# https://stackoverflow.com/questions/2187269/print-only-the-message-on-warnings
def _warning_setup():
    # pylint: disable=line-too-long
    formatwarning_orig = warnings.formatwarning
    warnings.formatwarning = lambda message, category, filename, lineno, line=None: \
        formatwarning_orig(message, category, filename, lineno, line='')



"""
Post-processing fairness intervension by choosing thresholds.

There are multiple definitions for choosing the thresholds:

1. Single threshold for all the sensitive attribute values
   that minimizes cost.
2. A threshold for each sensitive attribute value
   that minimize cost.
3. A threshold for each sensitive attribute value
   that achieve independence and minimize cost.
4. A threshold for each sensitive attribute value
   that achieve equal FNR (equal opportunity) and minimize cost.
5. A threshold for each sensitive attribute value
   that achieve separation (equalized odds) and minimize cost.

The code is based on `fairmlbook repository <https://github.com/fairmlbook/fairmlbook.github.io>`_.

References:
    - Hardt, M., Price, E., & Srebro, N. (2016).
      `Equality of opportunity in supervised learning
      <https://arxiv.org/abs/1610.02413>`_
      In Advances in neural information processing systems
      (pp. 3315-3323).
    - `Attacking discrimination with
      smarter machine learning by Google
      <https://research.google.com/bigpicture/attacking-discrimination-in-ml/>`_.

"""

# pylint: disable=no-name-in-module,ungrouped-imports

from collections import Counter

import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.ticker import AutoMinorLocator
from scipy.spatial import Delaunay

from responsibly.fairness.metrics.score import roc_curve_by_attr
from responsibly.fairness.metrics.utils import _groupby_y_x_sens
from responsibly.fairness.metrics.visualization import plot_roc_curves


TRINARY_SEARCH_TOL = 1e-3


def _strictly_increasing(arr):
    return (np.diff(arr) >= 0).all()


def _titlify(text):
    text = text.replace('_', ' ').title()
    if text == 'Fnr':
        text = 'FNR'
    return text


def _ternary_search_float(f, left, right, tol):
    """Trinary search: minimize f(x) over [left, right], to within +/-tol in x.

    Works assuming f is quasiconvex.
    """

    while right - left > tol:
        left_third = (2 * left + right) / 3
        right_third = (left + 2 * right) / 3
        if f(left_third) < f(right_third):
            right = right_third
        else:
            left = left_third
    return (right + left) / 2


def _ternary_search_domain(f, domain):
    """Trinary search: minimize f(x) over a domain (sequence).

    Works assuming f is quasiconvex and domain is ascending sorted.

    BUGGY, DO NOT USE

    >>> arr = np.concatenate([np.arange(10, 2, -1), np.arange(2, 20)])
    >>> t1 = _ternary_search_domain(lambda t: arr[t], range(len(arr)))
    >>> t2 = np.argmin(arr)

    >>> assert t1 == t2
    >>> assert arr[t1] == arr[t2]
    """

    left = 0
    right = len(domain) - 1
    changed = True

    while changed and left != right:

        changed = False

        left_third = (2 * left + right) // 3
        right_third = (left + 2 * right) // 3

        if f(domain[left_third]) < f(domain[right_third]):
            right = right_third - 1
            changed = True
        else:
            left = left_third + 1
            changed = True

    return domain[(left + right) // 2]


def _cost_function(fpr, tpr, base_rate, cost_matrix):
    """Compute the cost of given (fpr, tpr).

    [[tn, fp], [fn, tp]]
    """

    fp = fpr * (1 - base_rate)
    tn = (1 - base_rate) - fp
    tp = tpr * base_rate
    fn = base_rate - tp

    conf_matrix = np.array([tn, fp, fn, tp])

    return (conf_matrix * np.array(cost_matrix).ravel()).sum()


def _extract_threshold(roc_curves):
    return next(iter(roc_curves.values()))[2]


def _first_index_above(arr, value):
    """Find the smallest index i for which array[i] > value.

    If no such value exists, return len(array).
    """

    assert _strictly_increasing(arr), (
        'arr should be stricktly increasing.')

    arr = np.array(arr)
    v = np.concatenate([arr > value, [1]])
    return np.argmax(v, axis=0)


def _calc_acceptance_rate(fpr, tpr, base_rate):
    return (fpr * (1 - base_rate)
            + tpr * base_rate)


def find_single_threshold(roc_curves, base_rates, proportions,
                          cost_matrix):
    """Compute single threshold that minimizes cost.

    :param roc_curves: Receiver operating characteristic (ROC)
                       by attribute.
    :type roc_curves: dict
    :param base_rates: Base rate by attribute.
    :type base_rates: dict
    :param proportions: Proportion of each attribute value.
    :type proportions: dict
    :param cost_matrix: Cost matrix by [[tn, fp], [fn, tp]].
    :type cost_matrix: sequence
    :return: Threshold, FPR and TPR by attribute and cost value.
    :rtype: tuple

    """

    def total_cost_function(index):
        total_cost = 0

        for group, roc in roc_curves.items():
            fpr = roc[0][index]
            tpr = roc[1][index]

            group_cost = _cost_function(fpr, tpr,
                                        base_rates[group], cost_matrix)

            group_cost *= proportions[group]

            total_cost += group_cost

        return -total_cost

    thresholds = _extract_threshold(roc_curves)

    cost_per_threshold = [total_cost_function(index)
                          for index in range(len(thresholds))]
    cutoff_index = np.argmin(cost_per_threshold)

    fpr_tpr = {group: (roc[0][cutoff_index], roc[1][cutoff_index])
               for group, roc in roc_curves.items()}

    cost = total_cost_function(cutoff_index)

    return thresholds[cutoff_index], fpr_tpr, cost


def find_min_cost_thresholds(roc_curves, base_rates, proportions, cost_matrix):
    """Compute thresholds by attribute values that minimize cost.

    :param roc_curves: Receiver operating characteristic (ROC)
                       by attribute.
    :type roc_curves: dict
    :param base_rates: Base rate by attribute.
    :type base_rates: dict
    :param proportions: Proportion of each attribute value.
    :type proportions: dict
    :param cost_matrix: Cost matrix by [[tn, fp], [fn, tp]].
    :type cost_matrix: sequence
    :return: Thresholds, FPR and TPR by attribute and cost value.
    :rtype: tuple

    """
    # pylint: disable=cell-var-from-loop

    cutoffs = {}
    fpr_tpr = {}

    cost = 0
    thresholds = _extract_threshold(roc_curves)

    for group, roc in roc_curves.items():
        def group_cost_function(index):
            fpr = roc[0][index]
            tpr = roc[1][index]
            return -_cost_function(fpr, tpr,
                                   base_rates[group], cost_matrix)

        cost_per_threshold = [group_cost_function(index)
                              for index in range(len(thresholds))]
        cutoff_index = np.argmin(cost_per_threshold)

        cutoffs[group] = thresholds[cutoff_index]

        fpr_tpr[group] = (roc[0][cutoff_index],
                          roc[1][cutoff_index])

        cost += group_cost_function(cutoff_index) * proportions[group]

    return cutoffs, fpr_tpr, cost


def get_acceptance_rate_indices(roc_curves, base_rates,
                                acceptance_rate_value):
    indices = {}

    for group, roc in roc_curves.items():
        # can be calculated outside the function
        acceptance_rates = _calc_acceptance_rate(fpr=roc[0],
                                                 tpr=roc[1],
                                                 base_rate=base_rates[group])

        index = _first_index_above(acceptance_rates,
                                   acceptance_rate_value)

        indices[group] = index

    return indices


def find_independence_thresholds(roc_curves, base_rates, proportions,
                                 cost_matrix):
    """Compute thresholds that achieve independence and minimize cost.

    :param roc_curves: Receiver operating characteristic (ROC)
                       by attribute.
    :type roc_curves: dict
    :param base_rates: Base rate by attribute.
    :type base_rates: dict
    :param proportions: Proportion of each attribute value.
    :type proportions: dict
    :param cost_matrix: Cost matrix by [[tn, fp], [fn, tp]].
    :type cost_matrix: sequence
    :return: Thresholds, FPR and TPR by attribute and cost value.
    :rtype: tuple

    """

    cutoffs = {}

    def total_cost_function(acceptance_rate_value):
        # todo: move demo here + multiple cost
        #       + refactor - use threshold to calculate
        #         acceptance_rate_value
        indices = get_acceptance_rate_indices(roc_curves, base_rates,
                                              acceptance_rate_value)

        total_cost = 0

        for group, roc in roc_curves.items():
            index = indices[group]

            fpr = roc[0][index]
            tpr = roc[1][index]

            group_cost = _cost_function(fpr, tpr,
                                        base_rates[group],
                                        cost_matrix)

            group_cost *= proportions[group]

            total_cost += group_cost

        return -total_cost

    acceptance_rate_min_cost = _ternary_search_float(total_cost_function,
                                                     0, 1, TRINARY_SEARCH_TOL)

    cost = total_cost_function(acceptance_rate_min_cost)

    threshold_indices = get_acceptance_rate_indices(roc_curves, base_rates,
                                                    acceptance_rate_min_cost)
    thresholds = _extract_threshold(roc_curves)

    cutoffs = {group: thresholds[threshold_index]
               for group, threshold_index
               in threshold_indices.items()}

    fpr_tpr = {group: (roc[0][threshold_indices[group]],
                       roc[1][threshold_indices[group]])
               for group, roc in roc_curves.items()}

    return cutoffs, fpr_tpr, cost, acceptance_rate_min_cost


def get_fnr_indices(roc_curves, fnr_value):
    indices = {}

    tpr_value = 1 - fnr_value

    for group, roc in roc_curves.items():
        tprs = roc[1]
        index = _first_index_above(tprs,
                                   tpr_value) - 1
        index = max(0, index)
        indices[group] = index

    return indices


def find_fnr_thresholds(roc_curves, base_rates, proportions,
                        cost_matrix):
    """Compute thresholds that achieve equal FNRs and minimize cost.

    Also known as **equal opportunity**.

    :param roc_curves: Receiver operating characteristic (ROC)
                       by attribute.
    :type roc_curves: dict
    :param base_rates: Base rate by attribute.
    :type base_rates: dict
    :param proportions: Proportion of each attribute value.
    :type proportions: dict
    :param cost_matrix: Cost matrix by [[tn, fp], [fn, tp]].
    :type cost_matrix: sequence
    :return: Thresholds, FPR and TPR by attribute and cost value.
    :rtype: tuple

    """

    cutoffs = {}

    def total_cost_function(fnr_value):
        # todo: move demo here + multiple cost
        indices = get_fnr_indices(roc_curves, fnr_value)

        total_cost = 0

        for group, roc in roc_curves.items():
            index = indices[group]

            fpr = roc[0][index]
            tpr = roc[1][index]

            group_cost = _cost_function(fpr, tpr,
                                        base_rates[group],
                                        cost_matrix)
            group_cost *= proportions[group]

            total_cost += group_cost

        return -total_cost

    fnr_value_min_cost = _ternary_search_float(total_cost_function,
                                               0, 1,
                                               TRINARY_SEARCH_TOL)

    threshold_indices = get_fnr_indices(roc_curves, fnr_value_min_cost)

    cost = total_cost_function(fnr_value_min_cost)

    fpr_tpr = {group: (roc[0][threshold_indices[group]],
                       roc[1][threshold_indices[group]])
               for group, roc in roc_curves.items()}

    thresholds = _extract_threshold(roc_curves)
    cutoffs = {group: thresholds[threshold_index]
               for group, threshold_index
               in threshold_indices.items()}

    return cutoffs, fpr_tpr, cost, fnr_value_min_cost


def _find_feasible_roc(roc_curves):
    polygons = [Delaunay(list(zip(fprs, tprs)))
                for group, (fprs, tprs, _) in roc_curves.items()]

    feasible_points = []

    for poly in polygons:
        for p in poly.points:

            if all(poly2.find_simplex(p) != -1 for poly2 in polygons):
                feasible_points.append(p)

    return np.array(feasible_points)


def find_separation_thresholds(roc_curves, base_rate, cost_matrix):
    """Compute thresholds that achieve separation and minimize cost.

    Also known as **equalized odds**.

    :param roc_curves: Receiver operating characteristic (ROC)
                       by attribute.
    :type roc_curves: dict
    :param base_rate: Overall base rate.
    :type base_rate: float
    :param cost_matrix: Cost matrix by [[tn, fp], [fn, tp]].
    :type cost_matrix: sequence
    :return: Thresholds, FPR and TPR by attribute and cost value.
    :rtype: tuple

    """

    feasible_points = _find_feasible_roc(roc_curves)

    cost, (best_fpr, best_tpr) = max((_cost_function(fpr, tpr, base_rate,
                                                     cost_matrix),
                                      (fpr, tpr))
                                     for fpr, tpr in feasible_points)
    cost = - cost

    return {}, {'': (best_fpr, best_tpr)}, cost


def find_thresholds(roc_curves, proportions, base_rate,
                    base_rates, cost_matrix,
                    with_single=True, with_min_cost=True,
                    with_independence=True, with_fnr=True,
                    with_separation=True):
    """Compute thresholds that achieve various criteria and minimize cost.

    :param roc_curves: Receiver operating characteristic (ROC)
                       by attribute.
    :type roc_curves: dict
    :param proportions: Proportion of each attribute value.
    :type proportions: dict
    :param base_rate: Overall base rate.
    :type base_rate: float
    :param base_rates: Base rate by attribute.
    :type base_rates: dict
    :param cost_matrix: Cost matrix by [[tn, fp], [fn, tp]].
    :type cost_matrix: sequence

    :param with_single: Compute single threshold.
    :type with_single: bool
    :param with_min_cost: Compute minimum cost thresholds.
    :type with_min_cost: bool
    :param with_independence: Compute independence thresholds.
    :type with_independence: bool
    :param with_fnr: Compute FNR thresholds.
    :type with_fnr: bool
    :param with_separation: Compute separation thresholds.
    :type with_separation: bool

    :return: Dictionary of threshold criteria,
             and for each criterion:
             thresholds, FPR and TPR by attribute and cost value.
    :rtype: dict

    """

    thresholds = {}

    if with_single:
        thresholds['single'] = find_single_threshold(roc_curves,
                                                     base_rates,
                                                     proportions,
                                                     cost_matrix)

    if with_min_cost:
        thresholds['min_cost'] = find_min_cost_thresholds(roc_curves,
                                                          base_rates,
                                                          proportions,
                                                          cost_matrix)

    if with_independence:
        thresholds['independence'] = find_independence_thresholds(roc_curves,
                                                                  base_rates,
                                                                  proportions,
                                                                  cost_matrix)

    if with_fnr:
        thresholds['fnr'] = find_fnr_thresholds(roc_curves,
                                                base_rates,
                                                proportions,
                                                cost_matrix)

    if with_separation:
        thresholds['separation'] = find_separation_thresholds(roc_curves,
                                                              base_rate,
                                                              cost_matrix)

    return thresholds


def find_thresholds_by_attr(y_true, y_score, x_sens,
                            cost_matrix,
                            with_single=True, with_min_cost=True,
                            with_independence=True, with_fnr=True,
                            with_separation=True,
                            pos_label=None, sample_weight=None,
                            drop_intermediate=False):
    """
    Compute thresholds that achieve various criteria and minimize cost.

    :param y_true: Binary ground truth (correct) target values.
    :param y_score: Estimated target score as returned by a classifier.
    :param x_sens: Sensitive attribute values corresponded to each
                   estimated target.
    :param cost_matrix: Cost matrix by [[tn, fp], [fn, tp]].
    :type cost_matrix: sequence

    :param pos_label: Label considered as positive and others
                      are considered negative.
    :param sample_weight: Sample weights.
    :param drop_intermediate: Whether to drop some suboptimal
                              thresholds which would not appear on
                              a plotted ROC curve.
                              This is useful in order to create
                              lighter ROC curves.

    :param with_single: Compute single threshold.
    :type with_single: bool
    :param with_min_cost: Compute minimum cost thresholds.
    :type with_min_cost: bool
    :param with_independence: Compute independence thresholds.
    :type with_independence: bool
    :param with_fnr: Compute FNR thresholds.
    :type with_fnr: bool
    :param with_separation: Compute separation thresholds.
    :type with_separation: bool

    :return: Dictionary of threshold criteria,
             and for each criterion:
             thresholds, FPR and TPR by attribute and cost value.
    :rtype: dict
    """
    # pylint: disable=too-many-locals

    roc_curves = roc_curve_by_attr(y_true, y_score, x_sens,
                                   pos_label, sample_weight,
                                   drop_intermediate)

    proportions = {value: count / len(x_sens)
                   for value, count in Counter(x_sens).items()}

    if pos_label is None:
        pos_label = 1

    base_rate = np.mean(y_true == pos_label)
    grouped = _groupby_y_x_sens(y_true, y_score, x_sens)

    base_rates = {x_sens_value: np.mean(group['y_true'] == pos_label)
                  for x_sens_value, group in grouped}

    thresholds_data = find_thresholds(roc_curves,
                                      proportions,
                                      base_rate,
                                      base_rates,
                                      cost_matrix,
                                      with_single, with_min_cost,
                                      with_independence, with_fnr,
                                      with_separation)

    return thresholds_data


def plot_roc_curves_thresholds(roc_curves, thresholds_data,
                               aucs=None,
                               title='ROC Curves by Attribute',
                               ax=None, figsize=None,
                               title_fontsize='large',
                               text_fontsize='medium'):
    """Generate the ROC curves by attribute with thresholds.

    Based on :func:`skplt.metrics.plot_roc`

    :param roc_curves: Receiver operating characteristic (ROC)
                       by attribute.
    :type roc_curves: dict
    :param thresholds_data: Thresholds by attribute from the
                            function
                            :func:`~responsibly.interventions
                            .threshold.find_thresholds`.
    :type thresholds_data: dict
    :param aucs: Area Under the ROC (AUC) by attribute.
    :type aucs: dict
    :param str title: Title of the generated plot.
    :param ax: The axes upon which to plot the curve.
               If `None`, the plot is drawn on a new set of axes.
    :param tuple figsize: Tuple denoting figure size of the plot
                          e.g. (6, 6).
    :param title_fontsize: Matplotlib-style fontsizes.
                          Use e.g. 'small', 'medium', 'large'
                          or integer-values.
    :param text_fontsize: Matplotlib-style fontsizes.
                          Use e.g. 'small', 'medium', 'large'
                          or integer-values.
    :return: The axes on which the plot was drawn.
    :rtype: :class:`matplotlib.axes.Axes`

    """

    ax = plot_roc_curves(roc_curves, aucs,
                         title, ax, figsize, title_fontsize, text_fontsize)

    MARKERS = ['o', '^', 'x', '+', 'p']

    for (name, data), marker in zip(thresholds_data.items(), MARKERS):
        label = _titlify(name)
        ax.scatter(*zip(*data[1].values()),
                   marker=marker, color='k', label=label,
                   zorder=float('inf'))

    plt.legend()

    return ax


def plot_fpt_tpr(roc_curves,
                 title='FPR-TPR Curves by Attribute',
                 ax=None, figsize=None,
                 title_fontsize='large', text_fontsize='medium'):
    """Generate FPR and TPR curves by thresholds and by attribute.

    Based on :func:`skplt.metrics.plot_roc`

    :param roc_curves: Receiver operating characteristic (ROC)
                       by attribute.
    :type roc_curves: dict
    :param str title: Title of the generated plot.
    :param ax: The axes upon which to plot the curve.
               If `None`, the plot is drawn on a new set of axes.
    :param tuple figsize: Tuple denoting figure size of the plot
                          e.g. (6, 6).
    :param title_fontsize: Matplotlib-style fontsizes.
                          Use e.g. 'small', 'medium', 'large'
                          or integer-values.
    :param text_fontsize: Matplotlib-style fontsizes.
                          Use e.g. 'small', 'medium', 'large'
                          or integer-values.
    :return: The axes on which the plot was drawn.
    :rtype: :class:`matplotlib.axes.Axes`

    """

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)  # pylint: disable=unused-variable

    ax.set_title(title, fontsize=title_fontsize)

    thresholds = _extract_threshold(roc_curves)

    prop_cycle = plt.rcParams['axes.prop_cycle']
    colors = prop_cycle.by_key()['color']

    for (group, roc), color in zip(roc_curves.items(), colors):
        plt.plot(thresholds, roc[0], '-',
                 label='{} - FPR'.format(group), color=color)
        plt.plot(thresholds, roc[1], '--',
                 label='{} - TPR'.format(group), color=color)
        plt.legend()

    ax.set_ylim([0.0, 1.05])
    ax.set_xlabel('Threshold', fontsize=text_fontsize)
    ax.set_ylabel('Probability', fontsize=text_fontsize)
    ax.tick_params(labelsize=text_fontsize)
    ax.legend(fontsize=text_fontsize)

    return ax


def plot_costs(thresholds_data,
               title='Cost by Threshold Strategy',
               ax=None, figsize=None,
               title_fontsize='large', text_fontsize='medium'):
    """Plot cost by threshold definition and by attribute.

    Based on :func:`skplt.metrics.plot_roc`

    :param thresholds_data: Thresholds by attribute from the
                            function
                            :func:`~responsibly.interventions
                            .threshold.find_thresholds`.
    :type thresholds_data: dict
    :param str title: Title of the generated plot.
    :param ax: The axes upon which to plot the curve.
               If `None`, the plot is drawn on a new set of axes.
    :param tuple figsize: Tuple denoting figure size of the plot
                          e.g. (6, 6).
    :param title_fontsize: Matplotlib-style fontsizes.
                          Use e.g. 'small', 'medium', 'large'
                          or integer-values.
    :param text_fontsize: Matplotlib-style fontsizes.
                          Use e.g. 'small', 'medium', 'large'
                          or integer-values.
    :return: The axes on which the plot was drawn.
    :rtype: :class:`matplotlib.axes.Axes`
    """

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)  # pylint: disable=unused-variable

    ax.set_title(title, fontsize=title_fontsize)

    costs = {_titlify(group): cost
             for group, (_, _, cost, *_) in thresholds_data.items()}

    (pd.Series(costs)
     .sort_values(ascending=False)
     .plot(kind='barh', ax=ax))

    ax.set_xlabel('Cost', fontsize=text_fontsize)
    ax.set_ylabel('Threshold', fontsize=text_fontsize)
    ax.tick_params(labelsize=text_fontsize)

    return ax


def plot_thresholds(thresholds_data,
                    markersize=7,
                    title='Thresholds by Strategy and Attribute',
                    xlim=None,
                    ax=None, figsize=None,
                    title_fontsize='large', text_fontsize='medium'):
    """Plot thresholds by strategy and by attribute.

    Based on :func:`skplt.metrics.plot_roc`

    :param thresholds_data: Thresholds by attribute from the
                            function
                            :func:`~responsibly.interventions
                            .threshold.find_thresholds`.
    :type thresholds_data: dict
    :param int markersize: Marker size.
    :param str title: Title of the generated plot.
    :param tuple xlim: Set the data limits for the x-axis.
    :param ax: The axes upon which to plot the curve.
               If `None`, the plot is drawn on a new set of axes.
    :param tuple figsize: Tuple denoting figure size of the plot
                          e.g. (6, 6).
    :param title_fontsize: Matplotlib-style fontsizes.
                          Use e.g. 'small', 'medium', 'large'
                          or integer-values.
    :param text_fontsize: Matplotlib-style fontsizes.
                          Use e.g. 'small', 'medium', 'large'
                          or integer-values.
    :return: The axes on which the plot was drawn.
    :rtype: :class:`matplotlib.axes.Axes`
    """

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)  # pylint: disable=unused-variable

    ax.set_title(title, fontsize=title_fontsize)

    # TODO: refactor!
    df = pd.DataFrame({_titlify(key): thresholds
                       for key, (thresholds, *_) in thresholds_data.items()
                       if key != 'separation'})
    melted_df = pd.melt(df, var_name='Strategy', value_name='Threshold')
    melted_df['Attribute'] = list(df.index) * len(df.columns)

    sns.stripplot(y='Strategy', x='Threshold', hue='Attribute', data=melted_df,
                  jitter=False, dodge=True, size=markersize, ax=ax)

    minor_locator = AutoMinorLocator(2)
    fig.gca().yaxis.set_minor_locator(minor_locator)
    ax.grid(which='minor')

    if xlim is not None:
        ax.set_xlim(*xlim)

    ax.set_xlabel('Threshold', fontsize=text_fontsize)
    ax.set_ylabel('Strategy', fontsize=text_fontsize)
    ax.tick_params(labelsize=text_fontsize)

    return ax



import abc


class Dataset(abc.ABC):
    """Base class for datasets.

    Attributes
        - `df` - :class:`pandas.DataFrame` that holds the actual data.

        - `target` - Column name of the variable to predict
                    (ground truth)

        - `sensitive_attributes` - Column name of the
                                sensitive attributes

        - `prediction` - Columns name of the
                        prediction (optional)

    """

    @abc.abstractmethod
    def __init__(self, target, sensitive_attributes, prediction=None):
        """Load, preprocess and validate the dataset.

        :param target: Column name of the variable
                    to predict (ground truth)
        :param sensitive_attributes: Column name of the
                                    sensitive attributes
        :param prediction: Columns name of the
                           prediction (optional)
        :type target: str
        :type sensitive_attributes: list
        :type prediction: str
        """

        self.df = self._load_data()

        self._preprocess()

        self._name = self.__doc__.splitlines()[0]

        self.target = target
        self.sensitive_attributes = sensitive_attributes
        self.prediction = prediction

        self._validate()

    def __str__(self):
        return ('<{} {} rows, {} columns'
                ' in which {{{}}} are sensitive attributes>'
                .format(self._name,
                        len(self.df),
                        len(self.df.columns),
                        ', '.join(self.sensitive_attributes)))

    @abc.abstractmethod
    def _load_data(self):
        pass

    @abc.abstractmethod
    def _preprocess(self):
        pass

    @abc.abstractmethod
    def _validate(self):
        # pylint: disable=line-too-long

        assert self.target in self.df.columns,\
            ('the target label \'{}\' should be in the columns'
             .format(self.target))

        assert all(attr in self.df.columns
                   for attr in self.sensitive_attributes),\
            ('the sensitive attributes {{{}}} should be in the columns'
             .format(','.join(attr for attr in self.sensitive_attributes
                              if attr not in self.df.columns)))

        # assert all(attr in SENSITIVE_ATTRIBUTES
        #           for attr in self.sensitive_attributes),\
        # ('the sensitive attributes {} can be only from {}.'  # noqa
        #  .format(self.sensitive_attributes, SENSITIVE_ATTRIBUTES))



"""
Compute WEAT score of a Word Embedding.

WEAT is a bias measurement method for word embedding,
which is inspired by the `IAT <https://en.wikipedia.org/wiki/Implicit-association_test>`_
(Implicit Association Test) for humans.
It measures the similarity between two sets of *target words*
(e.g., programmer, engineer, scientist, ... and nurse, teacher, librarian, ...)
and two sets of *attribute words* (e.g., man, male, ... and woman, female ...).
A p-value is calculated using a permutation-test.

Reference:
    - Caliskan, A., Bryson, J. J., & Narayanan, A. (2017).
      `Semantics derived automatically
      from language corpora contain human-like biases
      <http://opus.bath.ac.uk/55288/>`_.
      Science, 356(6334), 183-186.

.. important::
    The effect size and pvalue in the WEAT have
    entirely different meaning from those reported in IATs (original finding).
    Refer to the paper for more details.

Stimulus and original finding from:

- [0, 1, 2]
  A. G. Greenwald, D. E. McGhee, J. L. Schwartz,
  Measuring individual differences in implicit cognition:
  the implicit association test.,
  Journal of personality and social psychology 74, 1464 (1998).

- [3, 4]:
  M. Bertrand, S. Mullainathan, Are Emily and Greg more employable
  than Lakisha and Jamal? a field experiment on labor market discrimination,
  The American Economic Review 94, 991 (2004).

- [5, 6, 9]:
  B. A. Nosek, M. Banaji, A. G. Greenwald, Harvesting implicit group attitudes
  and beliefs from a demonstration web site.,
  Group Dynamics: Theory, Research, and Practice 6, 101 (2002).

- [7]:
  B. A. Nosek, M. R. Banaji, A. G. Greenwald, Math=male, me=female,
  therefore math≠me.,
  Journal of Personality and Social Psychology 83, 44 (2002).

- [8]
  P. D. Turney, P. Pantel, From frequency to meaning:
  Vector space models of semantics,
  Journal of Artificial Intelligence Research 37, 141 (2010).
"""

# pylint: disable=C0301

import copy
import random
import warnings

import numpy as np
import pandas as pd
from mlxtend.evaluate import permutation_test

from responsibly.consts import RANDOM_STATE
from responsibly.utils import _warning_setup
from responsibly.we.data import WEAT_DATA
from responsibly.we.utils import (
    assert_gensim_keyed_vectors, cosine_similarities_by_words,
)


FILTER_BY_OPTIONS = ['model', 'data']
RESULTS_DF_COLUMNS = ['Target words', 'Attrib. words',
                      'Nt', 'Na', 's', 'd', 'p']
PVALUE_METHODS = ['exact', 'approximate']
PVALUE_DEFUALT_METHOD = 'exact'
ORIGINAL_DF_COLUMNS = ['original_' + key for key in ['N', 'd', 'p']]
WEAT_WORD_SETS = ['first_target', 'second_target',
                  'first_attribute', 'second_attribute']
PVALUE_EXACT_WARNING_LEN = 10

_warning_setup()


def _calc_association_target_attributes(model, target_word,
                                        first_attribute_words,
                                        second_attribute_words):
    # pylint: disable=line-too-long

    assert_gensim_keyed_vectors(model)

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', FutureWarning)

        first_mean = (cosine_similarities_by_words(model,
                                                   target_word,
                                                   first_attribute_words)
                      .mean())

        second_mean = (cosine_similarities_by_words(model,
                                                    target_word,
                                                    second_attribute_words)
                       .mean())

    return first_mean - second_mean


def _calc_association_all_targets_attributes(model, target_words,
                                             first_attribute_words,
                                             second_attribute_words):
    return [_calc_association_target_attributes(model, target_word,
                                                first_attribute_words,
                                                second_attribute_words)
            for target_word in target_words]


def _calc_weat_score(model,
                     first_target_words, second_target_words,
                     first_attribute_words, second_attribute_words):

    (first_associations,
     second_associations) = _calc_weat_associations(model,
                                                    first_target_words,
                                                    second_target_words,
                                                    first_attribute_words,
                                                    second_attribute_words)

    return sum(first_associations) - sum(second_associations)


def _calc_weat_pvalue(first_associations, second_associations,
                      method=PVALUE_DEFUALT_METHOD):

    if method not in PVALUE_METHODS:
        raise ValueError('method should be one of {}, {} was given'.format(
            PVALUE_METHODS, method))

    pvalue = permutation_test(first_associations, second_associations,
                              func=lambda x, y: sum(x) - sum(y),
                              method=method,
                              seed=RANDOM_STATE)  # if exact - no meaning
    return pvalue


def _calc_weat_associations(model,
                            first_target_words, second_target_words,
                            first_attribute_words, second_attribute_words):

    assert len(first_target_words) == len(second_target_words)
    assert len(first_attribute_words) == len(second_attribute_words)

    first_associations = _calc_association_all_targets_attributes(model,
                                                                  first_target_words,
                                                                  first_attribute_words,
                                                                  second_attribute_words)

    second_associations = _calc_association_all_targets_attributes(model,
                                                                   second_target_words,
                                                                   first_attribute_words,
                                                                   second_attribute_words)

    return first_associations, second_associations


def _filter_by_data_weat_stimuli(stimuli):
    """Filter WEAT stimuli words if there is a `remove` key.

    Some of the words from Caliskan et al. (2017) seems as not being used.

    Modifiy the stimuli object in place.
    """

    for group in stimuli:
        if 'remove' in stimuli[group]:
            words_to_remove = stimuli[group]['remove']
            stimuli[group]['words'] = [word for word in stimuli[group]['words']
                                       if word not in words_to_remove]


def _sample_if_bigger(seq, length):
    random.seed(RANDOM_STATE)
    if len(seq) > length:
        seq = random.sample(seq, length)
    return seq


def _filter_by_model_weat_stimuli(stimuli, model):
    """Filter WEAT stimuli words if they are not exists in the model.

    Modifiy the stimuli object in place.
    """

    for group_category in ['target', 'attribute']:
        first_group = 'first_' + group_category
        second_group = 'second_' + group_category

        first_words = [word for word in stimuli[first_group]['words']
                       if word in model]
        second_words = [word for word in stimuli[second_group]['words']
                        if word in model]

        min_len = min(len(first_words), len(second_words))

        # sample to make the first and second word set
        # with equal size
        first_words = _sample_if_bigger(first_words, min_len)
        second_words = _sample_if_bigger(second_words, min_len)

        first_words.sort()
        second_words.sort()

        stimuli[first_group]['words'] = first_words
        stimuli[second_group]['words'] = second_words


def _filter_weat_data(weat_data, model, filter_by):
    """inplace."""

    if filter_by not in FILTER_BY_OPTIONS:
        raise ValueError('filter_by should be one of {}, {} was given'.format(
            FILTER_BY_OPTIONS, filter_by))

    if filter_by == 'data':
        for stimuli in weat_data:
            _filter_by_data_weat_stimuli(stimuli)

    elif filter_by == 'model':
        for stimuli in weat_data:
            _filter_by_model_weat_stimuli(stimuli, model)


def calc_single_weat(model,
                     first_target, second_target,
                     first_attribute, second_attribute,
                     with_pvalue=True, pvalue_kwargs=None):
    """
    Calc the WEAT result of a word embedding.

    :param model: Word embedding model of ``gensim.model.KeyedVectors``
    :param dict first_target: First target words list and its name
    :param dict second_target: Second target words list and its name
    :param dict first_attribute: First attribute words list and its name
    :param dict second_attribute: Second attribute words list and its name
    :param bool with_pvalue: Whether to calculate the p-value of the
                             WEAT score (might be computationally expensive)
    :return: WEAT result (score, size effect, Nt, Na and p-value)
    """

    if pvalue_kwargs is None:
        pvalue_kwargs = {}

    (first_associations,
     second_associations) = _calc_weat_associations(model,
                                                    first_target['words'],
                                                    second_target['words'],
                                                    first_attribute['words'],
                                                    second_attribute['words'])

    if first_associations and second_associations:
        score = sum(first_associations) - sum(second_associations)
        std_dev = np.std(first_associations + second_associations, ddof=0)
        effect_size = ((np.mean(first_associations) - np.mean(second_associations))
                       / std_dev)

        pvalue = None
        if with_pvalue:
            pvalue = _calc_weat_pvalue(first_associations,
                                       second_associations,
                                       **pvalue_kwargs)
    else:
        score, std_dev, effect_size, pvalue = None, None, None, None

    return {'Target words': '{} vs. {}'.format(first_target['name'],
                                               second_target['name']),
            'Attrib. words': '{} vs. {}'.format(first_attribute['name'],
                                                second_attribute['name']),
            's': score,
            'd': effect_size,
            'p': pvalue,
            'Nt': '{}x2'.format(len(first_target['words'])),
            'Na': '{}x2'.format(len(first_attribute['words']))}


def calc_weat_pleasant_unpleasant_attribute(model,
                                            first_target, second_target,
                                            with_pvalue=True, pvalue_kwargs=None):
    """
    Calc the WEAT result with pleasent vs. unpleasant attributes.

    :param model: Word embedding model of ``gensim.model.KeyedVectors``
    :param dict first_target: First target words list and its name
    :param dict second_target: Second target words list and its name
    :param bool with_pvalue: Whether to calculate the p-value of the
                             WEAT score (might be computationally expensive)
    :return: WEAT result (score, size effect, Nt, Na and p-value)
    """

    weat_data = {'first_attribute': copy.deepcopy(WEAT_DATA[0]['first_attribute']),
                 'second_attribute': copy.deepcopy(WEAT_DATA[0]['second_attribute']),
                 'first_target': first_target,
                 'second_target': second_target}

    _filter_by_model_weat_stimuli(weat_data, model)

    if pvalue_kwargs is None:
        pvalue_kwargs = {}

    return calc_single_weat(model,
                            **weat_data,
                            with_pvalue=with_pvalue, pvalue_kwargs=pvalue_kwargs)


def calc_all_weat(model, weat_data='caliskan', filter_by='model',
                  with_original_finding=False,
                  with_pvalue=True, pvalue_kwargs=None):
    """
    Calc the WEAT results of a word embedding on multiple cases.

    Note that for the effect size and pvalue in the WEAT have
    entirely different meaning from those reported in IATs (original finding).
    Refer to the paper for more details.

    :param model: Word embedding model of ``gensim.model.KeyedVectors``
    :param dict weat_data: WEAT cases data.
                           - If `'caliskan'` (default) then all
                              the experiments from the original will be used.
                           - If an interger, then the specific experiment by index
                             from the original paper will be used.
                           - If a interger, then tje specific experiments by indices
                             from the original paper will be used.

    :param bool filter_by: Whether to filter the word lists
                           by the `model` (`'model'`)
                           or by the `remove` key in `weat_data` (`'data'`).
    :param bool with_original_finding: Show the origina
    :param bool with_pvalue: Whether to calculate the p-value of the
                             WEAT results (might be computationally expensive)
    :return: :class:`pandas.DataFrame` of WEAT results
             (score, size effect, Nt, Na and p-value)
    """

    if weat_data == 'caliskan':
        weat_data = WEAT_DATA
    elif isinstance(weat_data, int):
        index = weat_data
        weat_data = WEAT_DATA[index:index + 1]
    elif isinstance(weat_data, tuple):
        weat_data = [WEAT_DATA[index] for index in weat_data]

    if (not pvalue_kwargs
            or pvalue_kwargs['method'] == PVALUE_DEFUALT_METHOD):
        max_word_set_len = max(len(stimuli[ws]['words'])
                               for stimuli in weat_data
                               for ws in WEAT_WORD_SETS)
        if max_word_set_len > PVALUE_EXACT_WARNING_LEN:
            warnings.warn('At least one stimuli has a word set bigger'
                          ' than {}, and the computation might take a while.'
                          ' Consider using \'exact\' as method'
                          ' for pvalue_kwargs.'.format(PVALUE_EXACT_WARNING_LEN))

    actual_weat_data = copy.deepcopy(weat_data)

    _filter_weat_data(actual_weat_data,
                      model,
                      filter_by)

    if weat_data != actual_weat_data:
        warnings.warn('Given weat_data was filterd by {}.'
                      .format(filter_by))

    results = []
    for stimuli in actual_weat_data:
        result = calc_single_weat(model,
                                  stimuli['first_target'],
                                  stimuli['second_target'],
                                  stimuli['first_attribute'],
                                  stimuli['second_attribute'],
                                  with_pvalue, pvalue_kwargs)

        # TODO: refactor - check before if one group is without words
        # because of the filtering
        if not all(group['words'] for group in stimuli.values()
                   if 'words' in group):
            result['score'] = None
            result['effect_size'] = None
            result['pvalue'] = None

        result['stimuli'] = stimuli

        if with_original_finding:
            result.update({'original_' + k: v
                           for k, v in stimuli['original_finding'].items()})
        results.append(result)

    results_df = pd.DataFrame(results)
    results_df = results_df.replace('nan', None)
    results_df = results_df.fillna('')

    # if not results_df.empty:
    cols = RESULTS_DF_COLUMNS[:]
    if with_original_finding:
        cols += ORIGINAL_DF_COLUMNS
    if not with_pvalue:
        cols.remove('p')
    else:
        results_df['p'] = results_df['p'].apply(lambda pvalue: '{:0.1e}'.format(pvalue)  # pylint: disable=W0108
                                                if pvalue else pvalue)

    results_df = results_df[cols]
    results_df = results_df.round(4)

    return results_df



import numpy as np


def assert_deep_almost_equal(expected, actual, *args, **kwargs):
    """
    Assert that two complex structures have almost equal contents.

    Compares lists, dicts and tuples recursively. Checks numeric values
    using test_case's :py:meth:`unittest.TestCase.assertAlmostEqual` and
    checks all other values with :py:meth:`unittest.TestCase.assertEqual`.
    Accepts additional positional and keyword arguments and pass those
    intact to assertAlmostEqual() (that's how you specify comparison
    precision).

    https://github.com/larsbutler/oq-engine/blob/master/tests/utils/helpers.py
    """
    # is_root = not '__trace' in kwargs
    # trace = kwargs.pop('__trace', 'ROOT')
    try:
        if isinstance(expected, (int, float, complex,
                                 np.float16, np.float32,
                                 np.float64)):
            np.testing.assert_allclose(expected, actual, *args, **kwargs)
        elif isinstance(expected, (list, tuple, np.ndarray)):
            assert len(expected) == len(actual)
            for index, _ in enumerate(expected):
                v1, v2 = expected[index], actual[index]
                assert_deep_almost_equal(v1, v2,
                                         # __trace=repr(index),
                                         *args, **kwargs)
        elif isinstance(expected, dict):
            assert set(expected) == set(actual)
            for key in expected:
                assert_deep_almost_equal(expected[key], actual[key],
                                         # __trace=repr(key),
                                         *args, **kwargs)
        else:
            assert expected == actual
    except AssertionError as exc:
        # exc.__dict__.setdefault('traces', []).append(trace)
        # if is_root:
        #     trace = ' -> '.join(reversed(exc.traces))
        #     exc = AssertionError('{}\nTRACE: {}'
        #                          .format(exc.message, trace))
        raise exc



"""Sample integration test module using pytest-describe and expecter."""
# pylint: disable=redefined-outer-name,unused-variable,expression-not-assigned,singleton-comparison

# import pytest
# from expecter import expect
#
# from click.testing import CliRunner
#
# from responsibly.cli import main
#
#
# @pytest.fixture
# def runner():
#     return CliRunner()
#
#
# def describe_cli():
#
#     def describe_conversion():
#
#         def when_integer(runner):
#             result = runner.invoke(main, ['42'])
#
#             expect(result.exit_code) == 0
#             expect(result.output) == "12.80165\n"
#
#         def when_invalid(runner):
#             result = runner.invoke(main, ['foobar'])
#
#             expect(result.exit_code) == 0
#             expect(result.output) == ""



import math

import gensim
import matplotlib.pylab as plt
import numpy as np
import pandas as pd
from six import string_types
from sklearn.cluster import KMeans
from sklearn.manifold import TSNE
from sklearn.metrics import accuracy_score


WORD_EMBEDDING_MODEL_TYPES = (gensim.models.keyedvectors.KeyedVectors,
                              gensim.models.keyedvectors.BaseKeyedVectors,
                              gensim.models.fasttext.FastText,
                              gensim.models.word2vec.Word2Vec,
                              gensim.models.base_any2vec.BaseWordEmbeddingsModel,)  # pylint: disable=line-too-long


def round_to_extreme(value, digits=2):
    place = 10**digits
    new_value = math.ceil(abs(value) * place) / place
    if value < 0:
        new_value = -new_value
    return new_value


def normalize(v):
    """Normalize a 1-D vector."""
    if v.ndim != 1:
        raise ValueError('v should be 1-D, {}-D was given'.format(
            v.ndim))
    norm = np.linalg.norm(v)
    if norm == 0:
        return v
    return v / norm


def cosine_similarity(v, u):
    """Calculate the cosine similarity between two vectors."""
    v_norm = np.linalg.norm(v)
    u_norm = np.linalg.norm(u)
    similarity = v @ u / (v_norm * u_norm)
    return similarity


def project_vector(v, u):
    """Projecting the vector v onto direction u."""
    normalize_u = normalize(u)
    return (v @ normalize_u) * normalize_u


def reject_vector(v, u):
    """Rejecting the vector v onto direction u."""
    return v - project_vector(v, u)


def project_reject_vector(v, u):
    """Projecting and rejecting the vector v onto direction u."""
    projected_vector = project_vector(v, u)
    rejected_vector = v - projected_vector
    return projected_vector, rejected_vector


def project_params(u, v):
    """Projecting and rejecting the vector v onto direction u with scalar."""
    normalize_u = normalize(u)
    projection = (v @ normalize_u)
    projected_vector = projection * normalize_u
    rejected_vector = v - projected_vector
    return projection, projected_vector, rejected_vector


def cosine_similarities_by_words(model, word, words):
    """Compute cosine similarities between a word and a set of other words."""

    assert isinstance(word, string_types), \
        'The arguemnt `word` should be a string.'
    assert not isinstance(words, string_types), \
        'The argument `words` should not be a string.'

    vec = model[word]
    vecs = [model[w] for w in words]
    return model.cosine_similarities(vec, vecs)


def update_word_vector(model, word, new_vector):
    model.vectors[model.vocab[word].index] = new_vector
    if model.vectors_norm is not None:
        model.vectors_norm[model.vocab[word].index] = normalize(new_vector)


def generate_one_word_forms(word):
    return [word.lower(), word.upper(), word.title()]


def generate_words_forms(words):
    return sum([generate_one_word_forms(word) for word in words], [])


def take_two_sides_extreme_sorted(df, n_extreme,
                                  part_column=None,
                                  head_value='',
                                  tail_value=''):
    head_df = df.head(n_extreme)[:]
    tail_df = df.tail(n_extreme)[:]

    if part_column is not None:
        head_df[part_column] = head_value
        tail_df[part_column] = tail_value

    return (pd.concat([head_df, tail_df])
            .drop_duplicates()
            .reset_index(drop=True))


def assert_gensim_keyed_vectors(model):
    if not isinstance(model, WORD_EMBEDDING_MODEL_TYPES):
        type_names = (model_type.__name__
                      for model_type in WORD_EMBEDDING_MODEL_TYPES)
        raise TypeError('model should be on of the types'
                        ' ({}), not {}.'
                        .format(', '.join(type_names),
                                type(model)))


def most_similar(model, positive=None, negative=None,
                 topn=10, restrict_vocab=None, indexer=None,
                 unrestricted=True):
    """
    Find the top-N most similar words.

    Positive words contribute positively towards the similarity,
    negative words negatively.

    This function computes cosine similarity between a simple mean
    of the projection weight vectors of the given words and
    the vectors for each word in the model.
    The function corresponds to the `word-analogy` and `distance`
    scripts in the original word2vec implementation.

    Based on Gensim implementation.

    :param model: Word embedding model of ``gensim.model.KeyedVectors``.
    :param list positive: List of words that contribute positively.
    :param list negative: List of words that contribute negatively.
    :param int topn: Number of top-N similar words to return.
    :param int restrict_vocab: Optional integer which limits the
                               range of vectors
                               which are searched for most-similar values.
                               For example, restrict_vocab=10000 would
                               only check the first 10000 word vectors
                               in the vocabulary order. (This may be
                               meaningful if you've sorted the vocabulary
                               by descending frequency.)
    :param bool unrestricted: Whether to restricted the most
                              similar words to be not from
                              the positive or negative word list.
    :return: Sequence of (word, similarity).
    """
    if topn is not None and topn < 1:
        return []

    if positive is None:
        positive = []
    if negative is None:
        negative = []

    model.init_sims()

    if (isinstance(positive, string_types)
            and not negative):
        # allow calls like most_similar('dog'),
        # as a shorthand for most_similar(['dog'])
        positive = [positive]

    if ((isinstance(positive, string_types) and negative)
            or (isinstance(negative, string_types) and positive)):
        raise ValueError('If positives and negatives are given, '
                         'both should be lists!')

    # add weights for each word, if not already present;
    # default to 1.0 for positive and -1.0 for negative words
    positive = [
        (word, 1.0) if isinstance(word, string_types + (np.ndarray,))
        else word
        for word in positive
    ]
    negative = [
        (word, -1.0) if isinstance(word, string_types + (np.ndarray,))
        else word
        for word in negative
    ]

    # compute the weighted average of all words
    all_words, mean = set(), []
    for word, weight in positive + negative:
        if isinstance(word, np.ndarray):
            mean.append(weight * word)
        else:
            mean.append(weight * model.word_vec(word, use_norm=True))
            if word in model.vocab:
                all_words.add(model.vocab[word].index)

    if not mean:
        raise ValueError("Cannot compute similarity with no input.")
    mean = gensim.matutils.unitvec(np.array(mean)
                                   .mean(axis=0)).astype(float)

    if indexer is not None:
        return indexer.most_similar(mean, topn)

    limited = (model.vectors_norm if restrict_vocab is None
               else model.vectors_norm[:restrict_vocab])
    dists = limited @ mean

    if topn is None:
        return dists

    best = gensim.matutils.argsort(dists,
                                   topn=topn + len(all_words),
                                   reverse=True)

    # if not unrestricted, then ignore (don't return)
    # words from the input
    result = [(model.index2word[sim], float(dists[sim]))
              for sim in best
              if unrestricted or sim not in all_words]

    return result[:topn]


def get_seed_vector(seed, bias_word_embedding):

    if seed == 'direction':
        positive_end = bias_word_embedding.positive_end
        negative_end = bias_word_embedding.negative_end
        bias_word_embedding._is_direction_identified()  # pylint: disable=protected-access
        seed_vector = bias_word_embedding.direction
    else:
        if seed == 'ends':
            positive_end = bias_word_embedding.positive_end
            negative_end = bias_word_embedding.negative_end

        else:
            positive_end, negative_end = seed

        seed_vector = normalize(bias_word_embedding.model[positive_end]
                                - bias_word_embedding.model[negative_end])

    return seed_vector, positive_end, negative_end


def plot_clustering_as_classification(X, y_true, random_state=1, ax=None):

    if ax is None:
        _, ax = plt.subplots(figsize=(10, 5))

    y_cluster = (KMeans(n_clusters=2, random_state=random_state)
                 .fit_predict(X))

    embedded_vectors = (TSNE(n_components=2, random_state=random_state)
                        .fit_transform(X))

    for y_value in np.unique(y_cluster):
        mask = (y_cluster == y_value)
        label = 'Positive' if y_value else 'Negative'
        ax.scatter(embedded_vectors[mask, 0],
                   embedded_vectors[mask, 1],
                   label=label)

    ax.legend()

    acc = accuracy_score(y_true, y_cluster)

    return max(acc, 1 - acc)
